import logging
import re

from ..base import ensure_list
from .decorator import check_null
from .util import drop_repeat
from .util import get_shortest_element


def extract_city(string: str, na=None):
  """从字符串中提取城市（可能包含县级市）"""
  logging.warning(
      "extract_city is deprecated, use mdt_biz_util.util.district.extract_city instead",
      DeprecationWarning
  )
  from .district import extract_city as _
  return _(string, na)


def get_single_list(start, length, step):
  """
  获取单个起点的位置端点列表，如果起始位置不是零，则会返回0到起始位置的列表
  Args:
    start: 起始位置
    length: 总长度
    step: 步长
  Examples:
    >>> get_single_list(0, 7, 3) == [[0, 3], [3, 6], [6, 7]]
    >>> get_single_list(1, 7, 3) == [[0, 1], [1, 4], [4, 7]]
  """
  if start > 0:
    res = [[0, start]]
  else:
    res = []
  for i in range(start, length, step):
    if i + step <= length:
      res.append([i, i + step])
    else:
      res.append([i, length])
  return res


def get_breaks(length, step):
  """
  通过长度和步长获所有可能的取端点列表的列表
  Examples:
    >>> get_breaks(10, 3) == [
    >>>   [[0, 3], [3, 6], [6, 9], [9, 10]],
    >>>   [[0, 1], [1, 4], [4, 7], [7, 10]],
    >>>   [[0, 2], [2, 5], [5, 8], [8, 10]]
    >>> ]
  """
  res = []
  for start in range(0, step):
    res.append(get_single_list(start, length, step))
  return res


def get_list_by_position(string: str, breaks: list):
  """
  按照位置信息将字符串拆解为多个字符串的列表
  Args:
    string: 需要拆分的字符串
    breaks: 位置集合列表的列表（左闭右开），[[1, 5], [5, 10], [10, 15]]
  """
  res = []
  for brk in breaks:
    res.append(string[brk[0]:brk[1]])
  return res


def join_unique(strs: list):
  """删除列表中连续重复的元素，并拼接成字符串"""
  strs = drop_repeat(strs)
  return ''.join(strs)


def drop_repeat_string_by_step(string, step):
  """根据指定的小块长度，获取去重后的字符串"""
  _list = []
  length = len(string)
  for breaks in get_breaks(length, step):
    res = get_list_by_position(string, breaks)
    str_drop = join_unique(res)
    _list.append(str_drop)

  return get_shortest_element(_list)


@check_null()
def drop_repeat_string(string,
                       min_length=3,
                       max_length=None):
  """
  删除连续重复的字符串，按照step从大到小删除重复字符，返回去重后最短的字符串

  Args:
    string: 要处理的字符串
    min_length: 识别的最短长度，默认为3
    max_length: 识别的最长长度，默认不限制

  Examples:
    >>> drop_repeat_string('上海市上海市杨浦区')
    '上海市杨浦区'
  """
  if len(string) < min_length * 2:
    return string
  _list = []
  if not max_length:
    max_length = len(string) // 2
  for step in range(max_length, min_length - 1, -1):
    str_drop = drop_repeat_string_by_step(string, step)
    _list.append(str_drop)
    string = get_shortest_element(_list)
  return string


@check_null(default_rv=[])
def easy_split(string: str, seps: list = None, extra_seps: list = None):
  """
  使用常见的分隔符将字符串拆分为列表

  Args:
    string:
    seps: 自定义分隔符
    extra_seps: 额外的分隔符
  """
  if '\\' in string:
    raise Exception('字符串中不能包含"\\"')
  if not seps:
    seps = [
      '，', ',',
      '、', '/',
      '；', ';',
      '。', '\\|', ' ',
    ]
  if extra_seps:
    seps.extend(extra_seps)
  seps = ensure_list(seps)
  return [i for i in re.split('|'.join(seps), string) if i]


def punctuation_en2cn(text: str):
  """将英文标点符号替换为中文标点符号"""
  if not text or text == '' or not isinstance(text, str):
    return

  replace_dict = {
    ',': '，',
    '.': '。',
    '?': '？',
    '!': '！',
    ';': '；',
    ':': '：',
    '(': '（',
    ')': '）',
    '[': '【',
    ']': '】',
    '{': '｛',
    '}': '｝',
    '"': '“',
  }
  pattern = re.compile(
      '|'.join(re.escape(key) for key in replace_dict.keys())
  )

  def _replace(match):
    return replace_dict[match.group(0)]

  return pattern.sub(_replace, text)


def cyclic_slice(s, n=5):
  """迭代生成长度为n的字符串"""
  length = len(s)
  for i in range(0, length - n + 1):
    yield s[i:i + n]


def _is_seq(_str):
  """判断一个字符串是否全部是递增或递减1的数字组成的，如‘1234’、‘876’"""
  # 初始化为递增或递减
  diff_ls = [1, -1]
  s0 = _str[0]
  for s in _str[1:]:
    _d = int(s) - int(s0)
    if _d not in diff_ls:
      return False
    # 根据前面的字符再次确定递增或递减
    diff_ls = [_d]
    s0 = s
  return True


def is_seq(num_str: str, n: int):
  """判断一个全部为数字的字符串是否含有连续n个递增或递减1的数字"""
  num_str = str(num_str)
  assert num_str.isdigit(), '输入值必须为字符串且全部由数字组成'
  length = len(num_str)
  if length < n:
    return False
  for s in cyclic_slice(num_str, n):
    if _is_seq(s):
      return True
  return False


def is_repeated(string, min_length: int = 1):
  """检查输入是重复字符串组成的字符串"""
  n = len(string)
  if min_length * 2 > n:
    return False
  for i in range(min_length, n // 2 + 1):
    if n % i == 0:
      substring = string[:i]
      if substring * (n // i) == string:
        return True
  return False



from .decorator import timer


def filter_by_dist(ind, dist, r):
  """通过距离列表筛选出距离小于r的点的位置"""
  outer = []
  for i, ds in enumerate(dist):
    inner = []
    for j, d in enumerate(ds):
      if d <= r:
        inner.append(ind[i][j])
    outer.append(inner)
  return outer


def filter_by_limit(ls, n: int):
  """筛选列表中的列表的前limit个元素"""
  return [i[:n] for i in ls]


@timer()
def kdtree_query_radius(xy_tree, xy_query, r, limit: int = None, leaf_size=2):
  """
  使用kdtree查询半径为r的点

  Args:
    xy_tree: 要构造数的点集
    xy_query: 查询的点集
    r: 查询半径
    limit: 数量限制
    leaf_size: kdtree叶子节点的大小
  """
  from sklearn.neighbors import KDTree
  tree = KDTree(xy_tree, leaf_size=leaf_size)
  ind, dist = tree.query_radius(xy_query, r=r,
                                return_distance=True,
                                sort_results=True)
  ind = [list(i) for i in ind]
  if limit:
    ind = filter_by_limit(ind, limit)
    dist = filter_by_limit(dist, limit)
  return ind, dist


@timer()
def kdtree_query(xy_tree, xy_query, limit: int = None, r=None, leaf_size=2):
  """
  使用kdtree查询最近的一个或多个点

  Args:
    xy_tree: 要构造数的点集
    xy_query: 查询的点集
    limit: 数量限制
    r: 半径限制
    leaf_size: kdtree叶子节点的大小
  """
  from sklearn.neighbors import KDTree
  tree = KDTree(xy_tree, leaf_size=leaf_size)
  dist, ind = tree.query(
      xy_query,
      k=limit or 1,
      return_distance=True)
  ind = [list(i) for i in ind]
  dist = [list(i) for i in dist]
  if r:
    ind = filter_by_dist(ind, dist, r)
    dist = filter_by_dist(dist, dist, r)
  return ind, dist


def kdtree_nearest(xy_tree,
                   xy_query,
                   limit: int = None,
                   r: float = None,
                   leaf_size=2):
  """使用kdtree查询最近点"""
  if limit or not (r or limit):
    return kdtree_query(
        xy_tree=xy_tree,
        xy_query=xy_query,
        limit=limit,
        r=r,
        leaf_size=leaf_size)
  return kdtree_query_radius(
      xy_tree=xy_tree,
      xy_query=xy_query,
      r=r,
      limit=limit,
      leaf_size=leaf_size)



import csv
import os
import warnings
from tempfile import NamedTemporaryFile

import pandas as pd
from shapely.geometry.base import BaseGeometry

from ..fs.oss import OssUtils
from ..geometry.df import auto2shapely
from ..geometry.df import shapely2wkt
from ..geometry.util import infer_geom_format
from ..geometry.util import wkb_dumps
from ..util.os import ensure_dirpath_exist
from ..util.os import extension
from ..util.os import split_path
from ..util.util import first_notnull_value
from . import ALL_EXTS
from .transformer import df_iter


def to_csv_by_line(data: (list, dict), filepath: str):
  """
  逐行写入csv文件

  Args:
    data: 要写入的数据列表或字典，如果传入的为字典，在缺少文件的情况下会先写入key为表头
    filepath: 文件名
  """
  if isinstance(data, list):
    _data = data
  elif isinstance(data, dict):
    _data = [*data.values()]
    if not os.path.exists(filepath):
      to_csv_by_line([*data.keys()], filepath)
  else:
    raise ValueError('data must be list or dict')

  with open(filepath, 'a') as f:
    csv_write = csv.writer(f, dialect='excel')
    csv_write.writerow(_data)


def to_sheets(data: dict, filepath: str, index=False):
  """
  将多个dataframe保存到不同的sheet中

  Args:
    data: 要保存的数据集，格式为：{sheet_name: DataFrame}
    filepath: 要保存的文件名
    index: 是否保存index
  """
  assert isinstance(data, dict), 'data must be dict'
  with pd.ExcelWriter(filepath) as writer:
    for sheet_name, _df in data.items():
      _df.to_excel(writer, sheet_name, index=index)


def to_parquet(df: pd.DataFrame, filepath: str, index=False):
  """保存parquet文件，默认将shapely对象转换为wkb格式"""
  shapely_columns = [
    c for c in df if isinstance(first_notnull_value(df[c]), BaseGeometry)
  ]
  for c in shapely_columns:
    df[c] = df[c].apply(wkb_dumps)
  df.to_parquet(filepath, index=index)


def to_file(
    df: pd.DataFrame,
    filepath: str,
    *,
    index=False,
    log=True,
    encoding=None,
    access_key=None,
    secret_key=None
):
  """
  根据文件扩展名，将Dataframe保存为文件

  Args:
    df: 要保存的Dataframe
    filepath: 文件路径，包含扩展名
    index: 是否保存索引，默认不保存
    log: 是否打印保存信息
    encoding: 保存文件的编码
    access_key: 阿里云OSS访问密钥
    secret_key: 阿里云OSS访问密钥
  """

  def _check_excel(_df, _ex):
    """检查excel文件是否超出限制"""
    if 'geometry' in _df:
      if infer_geom_format(_df['geometry']) == 'shapely':
        _df = shapely2wkt(_df)
      if (_df['geometry'].str.len() >= 32767).any():
        warnings.warn('Excel中geometry列的长度不能超过32767，该列可能被截断')
    if _ex == '.xlsx' and _df.shape[0] >= 1048576:
      warnings.warn('.xlsx中最多只能保存1048576行数据，该文件可能被截断')
    if _ex == '.xls' and _df.shape[0] >= 65536:
      warnings.warn('.xls中最多只能保存65536行数据，该文件可能被截断')
    return _df

  if filepath.startswith('oss://'):
    assert access_key and secret_key, 'access_key和secret_key不能为空'
    return to_oss(df, filepath, access_key, secret_key)

  ensure_dirpath_exist(filepath)
  if log:
    print(f'Saving: {filepath}, Rows：{df.shape[0]}')
  df = df.copy()
  ex = extension(filepath)
  assert ex in ALL_EXTS, f'不支持的文件扩展名：{ex}'

  if ex == '.csv':
    df.to_csv(filepath, index=index, encoding=encoding)
  if ex in ('.pa', '.parquet'):
    to_parquet(df, filepath, index=index)
  if ex in ('.xlsx', '.xls'):
    _check_excel(df, ex).to_excel(filepath, index=index)
  if ex == 'json':
    df.to_json(filepath, orient='records')
  if ex in ('.shp', '.geojson'):
    df = auto2shapely(df)
    df.to_file(filepath, encoding=encoding,
               driver='GeoJSON' if ex == '.geojson' else None)
  if ex in ('.sav', '.zsav'):
    import pyreadstat
    pyreadstat.write_sav(df, filepath)
  if ex == '.feather':
    df.to_feather(filepath, index=index)
  if ex == '.pickle':
    df.to_pickle(filepath)


def to_parts_file(df, dirpath,
                  chunksize=None,
                  parts=None,
                  to_ext='.csv',
                  **kwargs):
  """
  将Dataframe保存为多个文件

  Args:
    df: 要拆分保存的DataFrame
    dirpath: 文件保存的目录
    chunksize: 拆分保存的文件大小
    parts: 拆分保存的文件数量
    to_ext: 文件扩展名
  """
  for i, _df in enumerate(df_iter(df, chunksize=chunksize, parts=parts)):
    savefile = os.path.join(dirpath, f'part_{str(i).zfill(6)}{to_ext}')
    to_file(_df, savefile, **kwargs)


def to_oss(df: pd.DataFrame, filepath, access_key, secret_key, encoding=None):
  """写入oss"""
  assert filepath.startswith('oss://'), '文件路径必须以oss://开头'
  work_path, _name, ext = split_path(filepath)
  _oss = OssUtils(
      work_path=work_path,
      access_key=access_key,
      secret_key=secret_key
  )
  with NamedTemporaryFile(suffix=ext) as _temp:
    to_file(df, _temp.name, encoding=encoding)
    _oss.upload(_temp.name, filepath, overwrite=True)



from functools import cached_property
from functools import lru_cache

from ..base import ensure_list
from ..base import is_empty
from ..base import not_empty
from ..base import warn_
from ..resource.bd_region import cities
from ..resource.bd_region import get_bd_region
from ..resource.bd_region import regions
from ..resource.patterns import AddressPattern
from .decorator import check_null
from .util import re_fast


def if_many_value(ls, action='raise', warning=False):
  """
  处理多个值的情况
  Args:
    ls: 列表
    action: 处理方式，可选值：'raise', 'first', 'last', 'coerce', 'all'
    warning: 是否打印警告
  """
  ls = ensure_list(ls)
  if len(ls) == 1:
    return ls[0]
  if action == 'raise':
    raise ValueError(f'0个值多个值{ls}')
  if action == 'coerce':
    return
  if warning:
    warn_(f'匹配到0个值多个值：{ls}')
  if action == 'all':
    return ls
  if not ls:
    return
  if action == 'first':
    return ls[0]
  if action == 'last':
    return ls[-1]


@lru_cache()
def get_city_list():
  """全部城市（全程+简称）列表"""
  df = get_bd_region()[['城市名称', '城市简称']]
  city_list = [
    *df['城市名称'].unique().tolist(),
    *df['城市简称'].unique().tolist(),
  ]
  return [c for c in city_list if not_empty(c)]


def is_city(name):
  """是否是城市"""
  return name in get_city_list()


@lru_cache()
def get_region_list():
  """全部区县（全程+简称）列表"""
  df = get_bd_region()[['区县名称', '区县简称']]
  region_list = [
    *df['区县名称'].unique().tolist(),
    *df['区县简称'].unique().tolist(),
  ]
  return [c for c in region_list if not_empty(c)]


def is_region(name):
  """是否是区县"""
  return name in get_region_list()


@lru_cache()
def city_full_short_mapping():
  """全称和简称的映射关系"""
  df = get_bd_region()[['城市名称', '城市简称']]
  df = df.drop_duplicates(subset=['城市名称']).dropna()
  return df.set_index('城市名称').to_dict()['城市简称']


@lru_cache()
def city_short_full_mapping():
  """简称和全称的映射关系"""
  df = get_bd_region()[['城市名称', '城市简称']]
  df = df.drop_duplicates(subset=['城市简称']).dropna()
  return df.set_index('城市简称').to_dict()['城市名称']


@lru_cache()
def region_full_city_mapping():
  """区县名称和城市简称的映射关系，仅保留一对一的关系"""
  df = get_bd_region()[['区县名称', '城市简称']]
  df = df.drop_duplicates(subset=['区县名称'], keep=False).dropna()
  return df.set_index('区县名称').to_dict()['城市简称']


@lru_cache()
def region_short_city_mapping():
  """区县简称和城市简称的映射关系，仅保留一对一的关系"""
  df = get_bd_region()[['区县简称', '城市简称']]
  df = df.drop_duplicates(subset=['区县简称'], keep=False).dropna()
  return df.set_index('区县简称').to_dict()['城市简称']


@lru_cache()
def city_shortname_id_mapping():
  """城市名称和城市id的映射关系"""
  df = get_bd_region()[['城市简称', '城市id']]
  df = df.drop_duplicates(subset=['城市简称']).dropna()
  return df.set_index('城市简称').to_dict()['城市id']


@lru_cache()
def city_id_shortname_mapping():
  """城市id和城市名称的映射关系"""
  df = get_bd_region()[['城市id', '城市简称']]
  df = df.drop_duplicates(subset=['城市id']).dropna()
  return df.set_index('城市id').to_dict()['城市简称']


@lru_cache()
def norm_city_name(name, full=False, warning=False):
  """标准化城市名称为简称或全称"""
  if not name:
    return
  fs = city_full_short_mapping()
  sf = city_short_full_mapping()
  if full:
    city = name if name in fs else sf.get(name)
  else:
    city = name if name in sf else fs.get(name)
  if not city:
    warn_(f'未找到城市名称：{name}', if_or_not=warning)
  return city


@lru_cache()
def get_city_id_by_name(name: str, warning=True):
  """根据城市名称获取城市id"""
  name = norm_city_name(name, warning=warning)
  shortname_id = city_shortname_id_mapping()
  city_id = shortname_id.get(name)
  if not city_id:
    warn_(f'未找到城市id：{name}', if_or_not=warning)
  return city_id


@lru_cache()
def get_city_name_by_id(city_id: (str, int), full=False, warning=False):
  """根据城市id获取城市名称"""
  if isinstance(city_id, (int, float)):
    city_id = str(int(city_id))
  id_shortname = city_id_shortname_mapping()
  city_name = id_shortname.get(city_id)
  if not city_name:
    warn_(f'未找到城市名称：{city_id}', if_or_not=warning)
  return norm_city_name(city_name, full=full, warning=warning)


@lru_cache()
def ensure_city_name(name, full=False, warning=False):
  """输入城市或区县，返回城市"""
  if not name:
    return
  city = norm_city_name(name, full=full, warning=warning)
  if city:
    return city
  rf_c = region_full_city_mapping()
  rs_c = region_short_city_mapping()
  city = rf_c.get(name) or rs_c.get(name)
  if not city:
    warn_(f'未找到城市名称：{name}', if_or_not=warning)
  return norm_city_name(city, full=full, warning=warning)


def get_upload_region(df, city):
  """从bd_region表中提取某个城市的区县数据，并处理成可以上传的格式"""
  city_id = df[
    (df['short_name'] == city) &
    (df['level_type'] == 2)
    ]['id'].unique().tolist()[0]
  df = df[df['parent_id'] == city_id]
  df = df[['id', 'name', 'geometry']]
  df.columns = ['regioncode', 'region', 'geometry']
  df = df.reset_index(drop=True)
  df['regioncode'] = df['regioncode'].astype(int)
  print(df.shape)
  return df


def get_upload_street(df, city):
  """从bd_region表中提取某个城市的街道数据，并处理成可以上传的格式"""
  city_id = df[
    (df['short_name'] == city) &
    (df['level_type'] == 2)
    ]['id'].unique().tolist()[0]
  region_ids = df[df['parent_id'] == city_id]['id'].tolist()
  df = df[df['parent_id'].isin(region_ids)]
  df = df[['id', 'name', 'parent_id', 'geometry']]
  df.columns = ['towncode', 'town', 'regioncode', 'geometry']
  df = df.reset_index(drop=True)
  df['towncode'] = df['towncode'].astype(int)
  df['regioncode'] = df['regioncode'].astype(int)
  print(df.shape)
  return df


@check_null(default_rv=[])
def extract_possible_region(string: str) -> list:
  """从字符串中提取可能的城市、区县"""
  if not isinstance(string, str):
    return []
  ls = []
  city_num, region_num = 0, 0
  # 根据正则表达式提取城市和区县
  for pattern in [*AddressPattern.cities, *AddressPattern.regions]:
    if res := re_fast(pattern, string, warning=False):
      if is_city(res):
        ls.append(res)
        city_num += 1
      if is_region(res):
        ls.append(res)
        region_num += 1
  # 没提取出区县时，按照字符串匹配尝试提取
  if region_num == 0:
    for cr in regions():
      if cr in string:
        ls.append(cr)
        break
  # 没提取出城市时，按照字符串匹配尝试提取
  if city_num == 0:
    for cr in cities():
      if cr in string:
        ls.append(cr)
        break
  ls = list(set(ls))
  ls1 = [i for i in ls if i.endswith('市')]
  ls2 = [i for i in ls if not i.endswith('市')]
  return [*ls1, *ls2]


def get_city_and_region(string) -> tuple:
  """从字符串中提取城市、区县"""
  city_list, region_list = [], []
  ls = extract_possible_region(string)
  for i in ls:
    if is_city(i):
      city_list.append(i)
    if is_region(i):
      region_list.append(i)
  if region_list and not city_list:
    for r in region_list:
      if _city := ensure_city_name(r, warning=False):
        city_list.append(_city)
  return (
    city_list[0] if city_list else None,
    region_list[0] if region_list else None,
    city_list,
    region_list
  )


@lru_cache()
def extract_city(string: str, na=None):
  """从字符串中提取城市（可能包含县级市）"""
  res = get_city_and_region(string)
  rv = res[0] or res[1]
  if is_empty(rv):
    return na
  return rv


class District:
  def __init__(self):
    """自动转换行政区划"""
    self.df = get_bd_region()

  @cached_property
  def city_list(self):
    """全部城市（全程+简称）列表"""
    return get_city_list()

  @cached_property
  def region_list(self):
    """全部区县（全程+简称）列表"""
    return get_region_list()

  @staticmethod
  def is_city(name: str) -> bool:
    """判断是否是地级市"""
    return is_city(name)

  @staticmethod
  def is_region(name: str) -> bool:
    """判断是否是县市区"""
    return is_region(name)

  def city_names(self, name: str):
    """返回匹配到的全部城市名称列表"""
    res = []
    for c in ['城市名称', '城市简称', '区县名称', '区县简称']:
      if name in self.df[c].unique():
        res = self.df[self.df[c] == name]['城市名称'].unique().tolist()
        break
    return res

  def province_names(self, name: str):
    """返回匹配到的全部省份名称列表"""
    res = []
    for c in ['省份名称', '省份简称', '城市名称', '城市简称', '区县名称',
              '区县简称']:
      if name in self.df[c].unique():
        res = self.df[self.df[c] == name]['省份名称'].unique().tolist()
        break
    return res

  def _get_district(self, name: str,
                    type_: str,
                    if_not_unique: str = 'coerce',
                    warning: bool = True):
    """
    通过地名（城市、区县）获取所在的城市名称

    Args:
      name: 城市名称
      type_: 城市、区县，可选参数city、province
      if_not_unique: 当查询到多个结果如何处理，
        - None， 默认， 返回空值
        - 'first'， 返回第一个
        - 'last'， 返回最后一个
        - 'all'， 返回所有城市的列表
      warning: 是否打印警告
    """
    assert if_not_unique in ['coerce', 'first', 'last', 'all']
    if type_ == 'city':
      ls = self.city_names(name)
    elif type_ == 'province':
      ls = self.province_names(name)
    else:
      raise ValueError('type_ must be "city" or "province"')
    return if_many_value(ls, action=if_not_unique, warning=warning)

  def city(self, name: str,
           if_not_unique: str = 'coerce',
           warning: bool = True):
    """
    通过地名（城市、区县）获取所在的城市名称

    Args:
      name: 城市名称
      if_not_unique: 当查询到多个结果如何处理，
        - None， 默认， 返回空值
        - 'first'， 返回第一个
        - 'last'， 返回最后一个
        - 'all'， 返回所有城市的列表
      warning: 是否打印警告
    """
    return self._get_district(
        name=name,
        type_='city',
        if_not_unique=if_not_unique,
        warning=warning
    )

  def province(self, name, if_not_unique='coerce', warning=True):
    """
    通过地名（省份、城市、区县）获取所在的省份名称

    Args:
      name: 地名
      if_not_unique: 当查询到多个结果如何处理，
        - `None`， 默认， 返回空值
        - `first`， 返回第一个
        - `last`， 返回最后一个
        - 'all'， 返回所有省份的列表
      warning: 是否打印警告
    """
    return self._get_district(
        name=name,
        type_='province',
        if_not_unique=if_not_unique,
        warning=warning
    )

  @staticmethod
  def get_city_id_by_name(name: str, warning=True):
    """通过城市名称获取城市id"""
    return get_city_id_by_name(name, warning=warning)

  @staticmethod
  def get_city_name_by_id(city_id: int, full=False, warning=False):
    """通过城市id获取城市名称"""
    return get_city_name_by_id(city_id, full=full, warning=warning)



import logging
import uuid
import warnings
from functools import lru_cache

import pandas as pd
from tqdm import tqdm

from ..base import not_empty
from ..etl.transformer import create_columns
from ..util.decorator import check_null
from .amap import get_address_amap
from .amap import get_place_amap
from .baidu import get_address_baidu
from .baidu import get_place_baidu
from .util import DEFAULT_RES


@lru_cache(maxsize=1024)
def geocode(*,
            address,
            city,
            source,
            srs='wgs84',
            key_baidu=None,
            key_amap=None):
  """
  对单个地址进行地理编码的方法汇总，返回经纬度坐标，可通过参数选择geocoding方法

  Args:
    address: 地址或项目名称关键词
    city: 城市
    source: 接口选择。baidu/amap：百度或高德的地理编码接口，baidu_poi/amap_poi：百度或高德的地点检索接口
    srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
    key_baidu: 百度接口的key，公共key失效后可自行传入
    key_amap: 高德接口的key，公共key失效后可自行传入
  """
  assert source in ('baidu', 'baidu_poi', 'amap', 'amap_poi')
  if source == 'baidu':
    return get_address_baidu(address=address, city=city, srs=srs, key=key_baidu)
  if source == 'baidu_poi':
    return get_place_baidu(address=address, city=city, srs=srs, key=key_baidu)
  if source == 'amap':
    return get_address_amap(address=address, city=city, srs=srs, key=key_amap)
  if source == 'amap_poi':
    return get_place_amap(address=address, city=city, srs=srs, key=key_amap)


@check_null(default_rv=DEFAULT_RES)
def geocode_best_poi(
    address,
    city,
    sig=80,
    address_geocode=False,
    srs='wgs84',
    key_baidu=None,
    key_amap=None):
  """
  获取最优的地点检索结果

  Args:
    address: POI名称
    city: 城市
    sig: 解析结果相似度，默认为80，范围0-100
    address_geocode: 是否使用地理编码补全，默认False
    srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
    key_baidu: 百度接口的key，公共key失效后可自行传入
    key_amap: 高德接口的key，公共key失效后可自行传入
  """
  try:
    ar = get_place_amap(address, city, srs=srs, key=key_amap)
    if ar['score'] >= 90:
      return ar
    br = get_place_baidu(address, city, srs=srs, key=key_baidu)
    if br['score'] >= sig or ar['score'] >= sig:
      return br if br['score'] > ar['score'] else ar
    if address_geocode:
      br = get_address_baidu(address, city, srs=srs, key=key_baidu)
      if br['score'] >= 90:
        return br
      return get_address_amap(address, city, srs=srs, key=key_amap)
  except Exception as e:
    warnings.warn(f'结果获取失败：【{city}】：【{address}】,{e}')
  return DEFAULT_RES


@check_null(default_rv=DEFAULT_RES)
def geocode_best_address(address, city, srs='wgs84', key_baidu=None,
                         key_amap=None):
  """获取最优的地理编码结果"""
  try:
    br = get_address_baidu(address, city, srs=srs, key=key_baidu)
    if br['score'] >= 90:
      return br
    return get_address_amap(address, city, srs=srs, key=key_amap)
  except Exception as e:
    warnings.warn(f'结果获取失败：【{city}】：【{address}】,{e}')
  return DEFAULT_RES


@lru_cache(maxsize=1024)
def geocode_best(
    address,
    city,
    address_type: str,
    sig: int = 80,
    srs: str = 'wgs84',
    address_geocode: bool = False,
    **kwargs):
  """
  获取最优的geocoding结果

  Args:
    address: 地址
    city: 城市
    address_type: 'poi' 或 'address'，不同的地理类型对应不同的接口，精确度不同
    sig: 得分
    srs: 坐标系
    address_geocode: 是否使用地址编码进行补充
    **kwargs:
  """
  assert address_type in ('poi', 'address'), 'address_type可选参数为poi或address'
  if address_type == 'poi':
    return geocode_best_poi(
        address, city, sig=sig, address_geocode=address_geocode, srs=srs,
        **kwargs
    )
  return geocode_best_address(address, city, srs=srs, **kwargs)


def geocode_df(df: pd.DataFrame,
               by: str,
               city: (str, list),
               address_type: str,
               sig: int = 80,
               address_geocode: bool = False,
               with_detail: bool = True,
               srs: str = 'wgs84',
               c_lng='lng',
               c_lat='lat',
               ignore_existing: bool = True,
               progress_bar: bool = True,
               **kwargs):
  """
  基于dataframe进行geocoding

  Args:
    df: 数据集
    by: 地址或名称列
    city: 城市，如为list表示城市列，如['城市']；如为str表示全部使用该城市,如 "上海"
    address_type: 'poi' 或 'address'，不同的地理类型对应不同的接口，精确度不同
    sig: 解析结果相似度，默认为80，范围为0-100，仅当address_type为poi时生效
    address_geocode: poi解析时是否使用地理编码补全，默认False，仅当address_type为poi时生效
    with_detail: 是否返回详情信息，包括返回值、得分、接口来源等
    srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
    c_lng: 经度列名，默认为 'lng'
    c_lat: 纬度列名，默认为 'lat'
    ignore_existing: 是否忽略已经存在经纬的行，默认为 TRUE，即保持原有的经纬度不变
    key_baidu: 百度接口的key，公共key失效后可自行传入
    key_amap: 高德接口的key，公共key失效后可自行传入
    progress_bar: 是否显示进度条，默认为TRUE

  See Also:
    * :func:`ricco.geocode.geocode.geocode_best`
  """
  assert address_type in ('poi', 'address'), 'address_type可选参数为poi或address'
  assert isinstance(city, (str, list)), 'city参数类型应为 str 或 list'
  assert df.index.is_unique, 'df的index必须唯一'

  __c_city = f'city-{uuid.uuid1()}'
  # 最后要输出的列
  base_cols = [c_lng, c_lat]
  if with_detail:
    base_cols.extend(['rv', 'score', 'source'])
  # 补全列，方便后续进行update
  df = create_columns(df.copy(), base_cols)
  num_all = df[df[c_lng].isna()].shape[0]
  # 新增临时城市列
  df[__c_city] = city if isinstance(city, str) else df[city[0]]

  indexes = []
  datas = []
  for _data in tqdm(
      df[[__c_city, by, c_lng, c_lat]].itertuples(),
      desc='Geocoding', total=df.shape[0], disable=not progress_bar
  ):
    if ignore_existing and not_empty(_data[3]) and not_empty(_data[4]):
      # ignore_existing为TRUE时，已经存在经纬度的行不再进行geocoding
      continue
    rv = geocode_best(
        address=_data[2], city=_data[1], sig=sig,
        address_type=address_type, address_geocode=address_geocode,
        srs=srs, **kwargs
    )
    rv[c_lng] = rv['lng']
    rv[c_lat] = rv['lat']
    datas.append(rv)
    indexes.append(_data[0])
  df_temp = pd.DataFrame(datas, index=indexes)
  df.update(df_temp[base_cols])
  num_miss = df[df[c_lng].isna()].shape[0]
  logging.warning(f'总数：{num_all}，成功：{num_all - num_miss}，失败：{num_miss}')
  # 删除临时城市列
  del df[__c_city]
  return df



from ricco.util.district import District
from ricco.util.district import get_city_and_region
from ricco.util.district import norm_city_name


def test_district():
  ds = District()
  assert ds.city(None) == None
  assert ds.city('') == None
  assert ds.city('单县') == '菏泽市'
  assert ds.city('杨浦区') == '上海市'
  assert ds.province(None) == None
  assert ds.province('') == None
  assert ds.province('滕州市') == '山东省'


def test_norm_city_name():
  assert norm_city_name('上海', full=True) == '上海市'
  assert norm_city_name('上海市', full=True) == '上海市'
  assert norm_city_name('上海市', full=False) == '上海'
  assert norm_city_name('上海', full=False) == '上海'


def test_extract_city():
  s = '上海市杨浦区政立路111号'
  assert get_city_and_region(s) == ('上海市', '杨浦区', ['上海市'], ['杨浦区'])

  s = '浙江省杭州市滨江区西兴街道泰安路199号'
  assert get_city_and_region(s) == ('杭州市', '滨江区', ['杭州市'], ['滨江区'])

  s = '云南省红河哈尼族彝族自治州个旧市大屯镇团山村万城米兰春天25幢2号'
  assert get_city_and_region(s) == (
    '红河哈尼族彝族自治州', '个旧市', ['红河哈尼族彝族自治州'], ['个旧市']
  )

  s = '内蒙古乌兰察布市察哈尔右翼前旗平地泉北方汽车城B厅35号'
  assert get_city_and_region(s) == (
    '乌兰察布市', '察哈尔右翼前旗', ['乌兰察布市'], ['察哈尔右翼前旗']
  )

  s = '甘肃省武威市天祝藏族自治县东大滩乡圈湾村'
  assert get_city_and_region(s) == (
    '武威市', '天祝藏族自治县', ['武威市'], ['天祝藏族自治县']
  )

  s = '内蒙古自治区锡林郭勒盟锡林浩特市污水处理厂附近工业园区17号'
  assert get_city_and_region(s) == (
    '锡林郭勒盟', '锡林浩特市', ['锡林郭勒盟'], ['锡林浩特市']
  )

  s = '河北省石家庄市桥西区新石中路168号15栋2单元402室'
  assert get_city_and_region(s) == (
    '石家庄市', '桥西区', ['石家庄市'], ['桥西区']
  )



from functools import lru_cache

import pandas as pd

from . import P_POSTCODE


@lru_cache()
def get_postcode() -> pd.DataFrame:
  """获取邮编数据集；并进行缓存避免多次IO"""
  print('获取邮编数据集')
  df = pd.read_csv(
      P_POSTCODE,
      dtype=str,
      usecols=['province', 'city', 'county', 'postcode']
  )
  df = df.dropna(subset='postcode')
  df = df.drop_duplicates(subset='postcode', ignore_index=True)
  return df



import datetime as dt
import re
import warnings
from datetime import datetime
from datetime import timedelta
from functools import wraps

import pandas as pd
from dateutil.relativedelta import relativedelta

from .decorator import check_null
from .decorator import check_str


def to_str(func):
  """将日期转为字符串的装饰器"""

  @wraps(func)
  def wrapper(self):
    if self.dst_format:
      return func(self).strftime(self.dst_format)
    else:
      return func(self)

  return wrapper


@check_str
def infer_format(string):
  """自动判断日期格式"""
  mapping = {
    # sep: 年月日
    '^\d{4}年([1-9]|0[1-9]|1[0-2])月([1-9]|0[1-9]|[12][0-9]|3[01])日([1-9]|0[1-9]|1[0-9]|2[04])时$': '%Y年%m月%d日%H时',
    '^\d{4}年([1-9]|0[1-9]|1[0-2])月([1-9]|0[1-9]|[12][0-9]|3[01])日$': '%Y年%m月%d日',
    '^\d{4}年([1-9]|0[1-9]|1[0-2])月$': '%Y年%m月',
    '^\d{4}年$': '%Y年',
    # sep: -
    '^\d{4}-([1-9]|0[1-9]|1[0-2])-([1-9]|0[1-9]|[12][0-9]|3[01])$': '%Y-%m-%d',
    '^\d{4}-([1-9]|0[1-9]|1[0-2])$': '%Y-%m',
    # sep: /
    '^\d{4}/([1-9]|0[1-9]|1[0-2])/([1-9]|0[1-9]|[12][0-9]|3[01])$': '%Y/%m',
    '^\d{4}/([1-9]|0[1-9]|1[0-2])$': '%Y/%m/%d',
    # sep: .
    '^\d{4}\.([1-9]|0[1-9]|1[0-2])\.([1-9]|0[1-9]|[12][0-9]|3[01])$': '%Y.%m.%d',
    '^\d{4}\.([1-9]|0[1-9]|1[0-2])$': '%Y.%m',
  }
  for _pattern, _format in mapping.items():
    if re.match(_pattern, string):
      return _format


@check_str
def auto2date(string, errors='ignore'):
  """自动检查格式并输出日期"""
  if string == '':
    return

  if _format := infer_format(string):
    return pd.to_datetime(string, format=_format, errors=errors)
  return pd.to_datetime(string, errors=errors)


def is_valid_date(string, na=False):
  """判断是否是一个有效的日期字符串"""
  if string in ['', None]:
    return na
  if auto2date(string):
    return True
  return False


@check_null()
def excel2date(dates, date_type='str'):
  """
  excel的数字样式时间格式转日期格式

  Args:
    dates: 传入excel的日期格式，或ymd的日期格式
    date_type: 返回日期类型，str or date
  """
  assert date_type in (
    'string', 'str', str, 'date', 'datetime'
  ), 'date_type参数错误，可选参数为str或date'

  dates = str(dates)

  if re.match('^\d{4,5}$', dates):
    from xlrd import xldate_as_datetime
    _date = xldate_as_datetime(int(dates), 0)
    if date_type in ('date', 'datetime'):
      return _date
    if date_type in ('string', 'str', str):
      return datetime.strftime(_date, '%Y-%m-%d')

  if _date := auto2date(dates):
    if date_type in ('date', 'datetime'):
      return _date
    if date_type in ('string', 'str', str):
      return datetime.strftime(_date, '%Y-%m-%d')


class DT:
  """
  DT日期类

  Args:
    date: 初始化支持字符串、datetime.date和datetime.datetime格式作为基准日期。不指定时，默认基准日期今天。
    format: 当初始化为字符串时，需要输入format指定日期字符串格式。
    dst_format: 可选。未指定时，类方法输出结果为datetime.date日期格式。指定dst_format时，输出结果为符合dst_format的字符串格式。
  """

  def __init__(self,
               date: (str, dt.date, dt.datetime) = None,
               format='%Y-%m-%d',
               dst_format=None):
    warnings.warn('DT2在版本0.5.38及更新版本中可以使用，详情见util.dt.DT2',
                  DeprecationWarning)
    if date:
      if isinstance(date, str):
        self.date = dt.datetime.strptime(date, format).date()
      elif isinstance(date, dt.datetime):
        self.date = date.date()
      elif isinstance(date, dt.date):
        self.date = date
      else:
        raise TypeError(
            '无法识别的类型，date只能为字符串、datetime.date或datetime.datetime类型')
    else:
      self.date = dt.date.today()
    self.dst_format = dst_format

  @property
  @to_str
  def today(self):
    """今天"""
    return self.date

  @property
  @to_str
  def tomorrow(self):
    """明天"""
    return self.date_move(days=1)

  @property
  @to_str
  def yesterday(self):
    """昨天"""
    return self.date_move(days=-1)

  @property
  @to_str
  def the_day_after_tomorrow(self):
    """后天"""
    return self.date_move(days=2)

  @property
  @to_str
  def the_day_before_yesterday(self):
    """前天"""
    return self.date_move(days=-2)

  @property
  @to_str
  def one_year_ago(self):
    """一年前"""
    return self.date_move(years=-1)

  @property
  @to_str
  def first_day_of_this_month(self):
    """本月第一天"""
    return self.date.replace(day=1)

  @property
  @to_str
  def last_day_of_this_month(self):
    """本月最后一天"""
    return self.date_move(months=1).replace(day=1) + relativedelta(days=-1)

  @property
  @to_str
  def day_half_year_ago(self):
    """半年前"""
    return self.date_move(months=-6)

  def date_move(self, years=0, months=0, days=0):
    """基于当前日期移动日期"""
    return self.date + relativedelta(years=years, months=months, days=days)

  @property
  @to_str
  def monday_of_this_week(self):
    """本周一"""
    weekday = self.date.isoweekday()
    if weekday == 1:
      return self.date
    return self.date - timedelta(weekday - 1)


class DT2:
  def __init__(self,
               date: (str, dt.date, dt.datetime) = None,
               date_format='%Y-%m-%d'):
    """
    DT日期类
    Args:
      date: 初始化支持字符串、datetime.date和datetime.datetime格式作为基准日期。不指定时，默认基准日期今天。
      date_format: 当初始化为字符串时，需要输入format指定日期字符串格式。
    Examples:
      today = '2024-01-31'
      上个月的第一天：
      DT(today).date_move(months=-1).first_day_of_this_month.get()
      --2023-12-01--(class 'datetime.date')
      DT(today).date_move(months=-1).first_day_of_this_month.get(format='%Y %m %d')
      --2023 12 01--(class 'str')
    """
    if date:
      if isinstance(date, str):
        self.date = dt.datetime.strptime(date, date_format).date()
      elif isinstance(date, dt.datetime):
        self.date = date.date()
      elif isinstance(date, dt.date):
        self.date = date
      else:
        raise TypeError(
            '无法识别的类型，date只能为字符串、datetime.date或datetime.datetime类型')
    else:
      self.date = dt.date.today()

  @property
  def today(self):
    """今天"""
    return self

  @property
  def tomorrow(self):
    """明天"""
    return self.date_move(days=1)

  @property
  def yesterday(self):
    """昨天"""
    return self.date_move(days=-1)

  @property
  def the_day_after_tomorrow(self):
    """后天"""
    return self.date_move(days=2)

  @property
  def the_day_before_yesterday(self):
    """前天"""
    return self.date_move(days=-2)

  @property
  def one_year_ago(self):
    """一年前的今天"""
    return self.date_move(years=-1)

  @property
  def first_day_of_this_month(self):
    """本月第一天"""
    self.date = self.date.replace(day=1)
    return self

  @property
  def last_day_of_this_month(self):
    """本月最后一天"""
    self.date = self.date_move(months=1).get().replace(day=1) + relativedelta(
        days=-1)
    return self

  @property
  def day_half_year_ago(self):
    """半年前的今天"""
    return self.date_move(months=-6)

  @property
  def monday_of_this_week(self):
    """本周的星期一"""
    weekday = self.date.isoweekday()
    if weekday == 1:
      return self
    self.date = self.date - dt.timedelta(weekday - 1)
    return self

  def date_move(self, years=0, months=0, days=0):
    """移动任意年/月/日后的日期，移动数量为正数表示向后移动，负数表示向前移动"""
    self.date = self.date + relativedelta(years=years, months=months, days=days)
    return self

  def get(self, format=None):
    """指定format，输出符合format格式的字符串，否则输出datetime.date类实例"""
    if format:
      return self.date.strftime(format)
    return self.date



import logging
import math
import warnings
from datetime import datetime
from functools import reduce
from itertools import product
from typing import List

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from pandas.tseries.offsets import MonthEnd

from ..base import ensure_list
from ..base import is_empty
from ..util.assertion import assert_not_null
from ..util.assertion import assert_series_unique
from ..util.decorator import check_null
from ..util.decorator import process_multi
from ..util.decorator import progress
from ..util.decorator import run_once
from ..util.decorator import timer
from ..util.util import and_
from ..util.util import fuzz_match
from ..util.util import or_
from ..util.util import sort_by_list
from ..util.util import to_float
from ..util.util import to_str_list
from .graph import get_graph_dict
from .graph import query_from_graph


def keep_best_unique(df: pd.DataFrame,
                     subset: (list, str),
                     value_cols: (str, list) = None
                     ) -> pd.DataFrame:
  """
  优化的去重函数，为保证数据的完整性，去重时优先去除指定列中的空值

  Args:
    df: 要去重的Dataframe
    subset: 按照哪些列去重
    value_cols: 优先去除那些列的空值，该列表是有顺序的
  """
  subset = ensure_list(subset)

  if not value_cols:
    value_cols = [i for i in df if i not in subset]
  value_cols = ensure_list(value_cols)

  return df.sort_values(
      value_cols, na_position='first'
  ).drop_duplicates(
      subset, keep='last'
  ).sort_index()


def table2dict(df: pd.DataFrame,
               key_col: str = None,
               value_col: (str, list) = None,
               orient: str = 'dict') -> dict:
  """
  DataFrame转字典

  Args:
    df:
    key_col: 生成key的列
    value_col: 生成value的列
    orient: 生成dict的方式，默认 'dict',还有 ‘list’, ‘series’, ‘split’, ‘records’, ‘index’
  """
  if not all([key_col, value_col]):
    columns = df.columns.tolist()
    key_col = columns[0]
    value_col = columns[1]
  df = df[df[key_col].notna()].set_index(key_col)
  if isinstance(value_col, list):
    df = df[value_col]
    return df.to_dict(orient=orient)
  else:
    df = df[[value_col]]
    return df.to_dict(orient=orient)[value_col]


def round_by_columns(df, columns: list):
  """对整列进行四舍五入，默认绝对值大于1的数值保留两位小数，小于1 的保留4位"""

  def _round(x):
    if is_empty(x):
      return np.nan
    if abs(x) >= 1:
      return round(x, 2)
    return round(x, 4)

  columns = ensure_list(columns)
  for c in columns:
    df[c] = df[c].apply(lambda x: _round(x))
  return df


def update_df(df: pd.DataFrame,
              new_df: pd.DataFrame,
              on: (str, list) = None,
              overwrite: bool = True,
              errors: str = 'ignore') -> pd.DataFrame:
  """
  根据某一列更新dataframe里的数据

  Args:
    df: 待升级的数据集
    new_df: 用于更新的DataFrame
    on: （可选参数）用于判断更新哪些行的列
    overwrite : （可选参数）控制如何处理原DataFrame在重叠位置上 **非空** 的值，默认为True

      * True: 默认值；使用 `other` DataFrame中的值覆盖原DataFrame中相应位置的值.
      * False: 只更新原DataFrame中重叠位置数据为 *空* 的值.
    errors: （可选参数）控制如何处理两个DataFrame同一位置都有值的行为，默认为 'ignore'

      * 'ignore': 默认值；DataFrame类型 df和other在同一个cell位置都是非NA值，使用other中的值替换df中的值。
      * 'raise': target和other都在同一位置包含非NA数据将抛出ValueError异常（'Data overlaps'）。
  """
  if on:
    df.set_index(on, inplace=True)
    df.update(new_df.set_index(on), overwrite=overwrite, errors=errors)
    df.reset_index(inplace=True)
  else:
    df.update(new_df, overwrite=overwrite, errors=errors)
  return df


def convert_date(df: pd.DataFrame,
                 columns: (str, list),
                 mode: str = 'first') -> pd.DataFrame:
  """
  将日期转为当月的第一天或最后一天

  Args:
    df: 要处理的DataFrame
    columns: 要转换的列
    mode: 'first' or 'last'
  """

  @check_null()
  def trans(x):
    return datetime(x.year, x.month, x.day)

  assert mode in ('first', 'last'), "可选参数为first or last"
  columns = ensure_list(columns)
  for c in columns:
    df[c] = pd.to_datetime(df[c])
    if mode == 'first':
      df[c] = df[c].apply(lambda x: x.replace(day=1))
    else:
      df[c] = pd.to_datetime(df[c], format="%Y%m") + MonthEnd(1)
    df[c] = df[c].apply(trans)
  return df


@timer()
@process_multi
def fuzz_df(df: pd.DataFrame,
            col: str,
            target_series: (list, pd.Series),
            c_dst: str = None,
            valid_score=60) -> pd.DataFrame:
  """
  模糊匹配。为DataFrame中的某一列从某个集合中模糊匹配匹配相似度最高的元素

  Args:
    df: 输入的dataframe
    col: 要匹配的列
    target_series: 从何处匹配， list/pd.Series
    c_dst: 关联后输出的列名，默认为原列名+"_target"后缀
    valid_score: 相似度大于该值的才返回，默认60
  """
  if c_dst:
    assert isinstance(c_dst, str), 'c_dst must be str'
  target_series = to_str_list(target_series)
  _df = df[df[col].notna()][[col]].drop_duplicates(ignore_index=True)

  c_dst = c_dst or f'{col}_target'
  _df[[
    c_dst, 'normal_score', 'partial_score', 'weight_score'
  ]] = _df.parallel_apply(
      lambda r: fuzz_match(r[col], target_series, valid_score=valid_score),
      result_type='expand',
      axis=1)
  return df.merge(_df, on=col, how='left')


def convert_to_float(df: pd.DataFrame,
                     columns: (list, str),
                     rex_method: str = 'mean') -> pd.DataFrame:
  """
  提取字符串中的数值信息并转为float类型

  Args:
    df: 要转换的DataFrame
    columns: 要转换的列
    rex_method: 计算mean,max,min， 默认为mean
  """
  columns = ensure_list(columns)
  for c in columns:
    df[c] = df[c].apply(lambda x: to_float(x, rex_method=rex_method))
  return df


def _is_eq(series: pd.Series, value):
  """整合空和非空的判断条件"""
  return series.isna() if pd.isna(value) else (series == value)


def _compare_dtypes(df_main, df_other):
  """判断df_main和df_other的列类型是否一致，其中df_main的列是df_other列的子集"""
  for c in df_main:
    if df_main[c].dtypes != df_other[c].dtypes:
      warnings.warn(f'两个数据集"{c}"列类型不一致')


def filter_by_df(df: pd.DataFrame,
                 df_sizer: pd.DataFrame,
                 reverse: bool = False) -> pd.DataFrame:
  """从df中筛选出满足df_size的数据"""
  _compare_dtypes(df_sizer, df)
  df_sizer = df_sizer.drop_duplicates()
  cond = or_(*[
    and_(
        *[_is_eq(df[c], df_sizer[c][i]) for c in df_sizer]
    ) for i in df_sizer.index
  ])
  if reverse:
    return df[~cond]
  return df[cond]


def drop_by_df(df: pd.DataFrame, df_deleted: pd.DataFrame) -> pd.DataFrame:
  """根据一个df_deleted中的数据从df中删除"""

  _compare_dtypes(df_deleted, df)
  df_deleted = df_deleted.drop_duplicates()

  return df[~or_(*[
    and_(
        *[_is_eq(df[c], df_deleted[c][i]) for c in df_deleted]
    ) for i in df_deleted.index
  ])]


def wrap_dict(df: pd.DataFrame, c_srcs: list = None, c_dst: str = 'extra'):
  """将指定的多列合并成一个字典列
  Args:
    df: 要处理的数据集
    c_srcs: 需要包装成字典的多列, 如果不指定，则wrap所有的列
    c_dst: 处理后的数据存储的列名称，默认为 'extra'
  """
  df = df.copy()
  if not c_srcs:
    c_srcs = df.columns.tolist()
  extra = df[c_srcs]
  df = df.drop(c_srcs, axis=1)
  df[c_dst] = extra.to_dict('records')
  return df


@timer()
def expand_dict(df: pd.DataFrame, c_src: str):
  """展开字典为多列"""
  df_extra = pd.json_normalize(
      df[c_src].apply(lambda x: {} if pd.isna(x) else x),
      max_level=0, errors='ignore',
  )
  df_extra.index = df.index
  return pd.concat([df.drop(columns=[c_src]), df_extra], axis=1)


@timer()
def expand_dict_silent(df: pd.DataFrame, c_src: str):
  """展开字典为多列"""
  logging.warning(
      DeprecationWarning,
      'expand_dict_silent is deprecated, use expand_dict instead'
  )
  return expand_dict(df, c_src)


@timer()
def split_to_rows(df: pd.DataFrame, column: str, delimiter: str = '|'):
  """
  含有多值的列分拆成多行

  Args:
    df:
    column: 存在多值的列
    delimiter: 多个值之间的分割符
  """
  df = df.copy()
  df[column] = df[column].str.split(delimiter)
  return df.explode(column, ignore_index=True)


@timer()
def split_list_to_row(df: pd.DataFrame, column):
  """将列表列中列表的元素拆成多行"""
  return df.explode(column, ignore_index=True)


def dict2df(data: dict, c_key='key', c_value='value', as_index=False):
  """将字典转为dataframe"""
  df = pd.DataFrame()
  df[c_key] = data.keys()
  df[c_value] = data.values()
  return df.set_index(c_key) if as_index else df


@timer()
def is_changed(df_old: pd.DataFrame,
               df_new: pd.DataFrame,
               key_cols: (str, list) = None,
               value_cols: (str, list) = None,
               c_res: str = 'is_changed') -> pd.DataFrame:
  """
  判断新旧数据集中的每一条数据是否变化

  Args:
    df_old: 原始数据集
    df_new: 修改后的数据集
    key_cols: 关键列，不指定则以索引列为准
    value_cols: 要对比的列，默认出key_cols之外的其他列
    c_res: 对比结果列名，默认为“is_changed”
  """
  # 数据集及参数检验
  key_cols = ensure_list(key_cols)
  assert_series_unique(df_new, key_cols)
  if not value_cols:
    error_msg = 'Each datasets must have same columns.'
    assert set(df_new.columns) == set(df_old.columns), error_msg
    value_cols = [c for c in df_old if c not in key_cols]
  # 对比
  # 默认所有的都是更改的
  df_new = df_new.copy()
  df_new[c_res] = 'Changed'
  df_temp = df_old[[*key_cols, *value_cols]].merge(
      df_new[[*key_cols, *value_cols, c_res]],
      how='left',
      on=key_cols)
  # 合并后为空的为 NotFound
  df_temp[c_res] = df_temp[c_res].fillna('NotFound')

  # 对比每一个值，如果每一列的每一个值都相同，则认为是没有变化
  cond = and_(
      *[
        (df_temp[f'{c}_x'] == df_temp[f'{c}_y']) |
        (df_temp[f'{c}_x'].isna() & df_temp[f'{c}_y'].isna())
        for c in value_cols
      ]
  )
  df_temp.loc[cond, c_res] = 'NotChange'
  df_temp = df_temp[[*key_cols, c_res]]
  return df_old.merge(df_temp, how='left', on=key_cols)


def df_iter(df: pd.DataFrame, *, chunksize: int = None, parts: int = None):
  """
  DataFrame切片生成器

  Args:
    df: 要切片的DataFrame
    chunksize: 每次返回的大小，与parts必须指定一个
    parts: 返回的次数，与chunksize必须指定一个
  """
  assert any([chunksize, parts]), 'chunksize和parts必须指定一个'
  size = df.shape[0]
  if chunksize:
    parts = math.ceil(size / chunksize)
  else:
    chunksize = math.ceil(size / parts)
  for i in range(parts):
    low = i * chunksize
    high = (i + 1) * chunksize
    if i == parts - 1:
      yield df.iloc[low:, :]
    else:
      yield df.iloc[low: high, :]


def create_columns(df: pd.DataFrame, columns: list, value=None):
  """
  创建新的列，默认为空

  Args:
    df: DataFrame
    columns: 列名
    value: 值，默认为空
  """
  columns = ensure_list(columns)
  for c in columns:
    if c not in df:
      df[c] = value
  return df


@progress
def expand_graph(df: pd.DataFrame,
                 start_level,
                 c_key='id',
                 c_level_type='level_type',
                 c_parent_key='parent_id',
                 c_info=None):
  """
  将图数据展开为宽表

  Args:
    df: 要展开的图数据
    start_level: 开始展开的层级，一般是最低层级
    c_key: 关键列
    c_level_type: 层级类型列
    c_parent_key: 父级关键列
    c_info: 要保留的其他列
  """

  @run_once
  def asserts(_df, c_id, c_type, c_parent_id):
    assert isinstance(_df, pd.DataFrame)
    assert _df[c_id].is_unique
    assert_not_null(_df, c_type), f'{c_type}不能为空'
    assert _df[_df[c_id] == _df[c_parent_id]].empty, '父子不能相同'

  assert start_level in df[c_level_type].unique(), 'start_level不在level_type中'
  df_base = df.copy()

  # 构造查询字典
  graph_df = df.copy()
  asserts(graph_df, c_id=c_key, c_type=c_level_type,
          c_parent_id=c_parent_key)
  graph_df[c_level_type] = graph_df[c_level_type].apply(
      lambda lp: f'{lp}_{c_key}'
  )
  # 事先将Dataframe转为字典，提高查询性能
  graph_dict = get_graph_dict(graph_df=graph_df, c_key=c_key,
                              c_level_type=c_level_type,
                              c_parent_key=c_parent_key,
                              c_parent_type='parent_type')
  # 将索引列提取出来作为基础底表
  df = df[df[c_level_type] == start_level]
  df = df[[c_key]].reset_index(drop=True)
  # 将索引列逐个传入查询
  df['extra'] = df[c_key].progress_apply(
      lambda x: query_from_graph(
          key=x, graph_data=graph_dict, c_key=c_key,
          c_level_type=c_level_type, c_parent_key=c_parent_key)
  )
  # 展开最终结果
  df = expand_dict(df, 'extra')
  if c_info:
    c_info = ensure_list(c_info)
    for level_type in df_base[c_level_type].unique():
      df_level = df_base[[c_key, *c_info]]
      df_level.columns = [f'{level_type}_{c}' for c in df_level]
      if f'{level_type}_{c_key}' in df_level:
        df = df.merge(df_level, on=f'{level_type}_{c_key}', how='left')
  return df


def null_if_in(df, columns: (str, list), values: list):
  """
  如果列的值正好在values中定义的数组中，则将该列值设置为None

  Args:
    df:
    columns: 需要检查的列，一个或多少
    values: 所有需要标识为None的值
  """
  columns = ensure_list(columns)
  for c in columns:
    conditions = [df[c] == v for v in values]
    df.loc[or_(*conditions), c] = None
  return df


def fix_null(df: pd.DataFrame, columns: (str, list) = None):
  """将None/NaN/NaT等都替换为None"""
  if not columns:
    columns = df.columns
  columns = ensure_list(columns)
  for c in columns:
    df.loc[df[c].isna(), c] = None
  return df


def fillna_by_adding(df: pd.DataFrame, col: str):
  """以自增的方式填充空值"""
  df = df.copy()
  max_id = 0 if df[col].isna().all() else df[col].max()
  n = df[col].isna().sum()
  df.loc[df[col].isna(), col] = range(int(max_id + 1), int(max_id + 1 + n))
  df[col] = df[col].astype(int)
  assert_not_null(df, col)
  assert_series_unique(df, col)
  return df


def rolling_by_month(df, c_date, c_group, months: int, agg: dict):
  """
  按照日期进行滚动聚合，比如：分组对某个项目计算近6个月的成交面积

  Args:
    df: 输入的dataframe
    c_date: 日期列
    c_group: 分组列
    months: 窗口大小
    agg: 聚合函数，如：{'金额': 'sum'}
  """
  df = df.copy()
  df[c_date] = pd.to_datetime(df[c_date])
  # 起始日期为最小日期加上窗口时间
  date_start = df[c_date].min() + relativedelta(months=months)
  date_end = df[c_date].max()
  dfs = []
  while date_end >= date_start:
    date_l = date_start - relativedelta(months=months)
    _df = df[(df[c_date] >= date_l) & (df[c_date] < date_start)]
    _df = _df.groupby(c_group, as_index=False).agg(agg)
    _df[c_date] = date_start
    dfs.append(_df.copy())
    date_start += relativedelta(months=1)
  return pd.concat(dfs, ignore_index=True)


def columns_contrast(df_left, df_right, n_rows=None):
  """对比两个Dataframe的列，同名的列会放在一起"""
  _l = 'left'
  _r = 'right'
  if n_rows:
    df_left = df_left.head(n_rows)
    df_right = df_right.head(n_rows)
  cols_left = df_left.columns.tolist()
  cols_right = df_right.columns.tolist()
  mapping_left = {i: f'{i}-{_l}' for i in cols_left}
  mapping_right = {i: f'{i}-{_r}' for i in cols_right}
  df_left = df_left.rename(columns=mapping_left)
  df_right = df_right.rename(columns=mapping_right)
  df = df_left.join(df_right)
  # 排序列
  common_list = [i for i in cols_left if i in cols_right]
  cols_new = [f'{i}-{j}' for i, j in product(common_list, [_l, _r])]
  cols_all = [*mapping_left.values(), *mapping_right.values()]
  cols_sorted = sort_by_list(cols_all, cols_new)
  return df[cols_sorted]


def merge_dfs(dfs: List[pd.DataFrame], on, how):
  """
  合并多个dataframe

  Args:
    dfs: 多个Dataframe组成的列表，需要有相同的列
    on: 根据那个字段进行merge
    how: 以何种方式进行合并，同pd.merge
  """
  return reduce(
      lambda left, right: pd.merge(left, right, on=on, how=how),
      dfs
  )


def shift_data(
    df,
    c_date: str,
    c_group: (str, list),
    c_value: (str, list),
    subfix: str = '_环比',
    shift: int = 1
):
  """
  环比计算

  Args:
    df: DataFrame
    c_date: 日期列
    c_group: 分组列
    c_value: 值列
    subfix: 环比列后缀
    shift: 环比计算的时间间隔
  """
  c_value = ensure_list(c_value)
  sort_keys = [*ensure_list(c_group), *ensure_list(c_date)]
  df = df.sort_values(by=sort_keys, ignore_index=True)
  for c in c_value:
    df[f'{c}_1'] = df.groupby(c_group)[c].shift(shift)
    df[f'{c}{subfix}'] = (
        (df[c] - df[f'{c}_1'])
        / df[f'{c}_1']
    )
    del df[f'{c}_1']
  return df



import re
import warnings

import pandas as pd

from ..base import ensure_list
from ..base import not_empty
from .util import and_
from .util import is_digit
from .util import is_unique_series


def skip_column(df: pd.DataFrame, col: str, skip_if_not_exists: bool = True):
  df = df.copy().reset_index(drop=True)
  if col not in df:
    if skip_if_not_exists:
      warnings.warn(f'"{col}"列不存在，已跳过')
      return True
    else:
      raise KeyError(f'"{col}"列不存在')
  return False


def assert_columns_exists(df: pd.DataFrame, columns: list):
  """检查列是否存在"""
  df = df.copy().reset_index(drop=True)
  columns = ensure_list(columns)
  cols = [c for c in columns if c not in df]
  assert not cols, f'缺失"{cols}"列'


def assert_not_empty_str(df: pd.DataFrame,
                         col: str,
                         skip_if_not_exists: bool = True):
  """校验是否存在空白字符串"""
  df = df.copy().reset_index(drop=True)
  if skip_column(df, col, skip_if_not_exists):
    return
  assert '' not in df[col].values, f'"{col}"列存在空白字符串'


def assert_not_null(df: pd.DataFrame,
                    col: str,
                    skip_if_not_exists: bool = True):
  """检查是否非空（空白字符串认为是空值）"""
  df = df.copy().reset_index(drop=True)
  if skip_column(df, col, skip_if_not_exists):
    return
  if not isinstance(col, str):
    raise TypeError('col参数仅支持str类型')
  if df[col].isna().any():
    raise AssertionError(f'"{col}"列存在空值')
  assert_not_empty_str(df, col)


def assert_values_in(df: pd.DataFrame,
                     col: str,
                     enums: (dict, list),
                     skip_if_not_exists: bool = True):
  """
  检查Dataframe中某一列的值是否在指定的值的范围内

  Args:
    df: 要检查的dataframe
    col: 列名
    enums: 指定的enum值，当传入dict时，包含在key和value中的值都通过
    skip_if_not_exists: 当列不存在时是否跳过
  """
  df = df.copy().reset_index(drop=True)
  if skip_column(df, col, skip_if_not_exists):
    return
  if isinstance(enums, dict):
    vs = [*list(enums.keys()), *list(enums.values())]
  elif isinstance(enums, list):
    vs = enums
  else:
    raise TypeError('enums类型错误，list or dict')
  rv = [t for t in df[col].unique().tolist() if not_empty(t) and t not in vs]
  assert not rv, f'{rv}不是"{col}"列有效的enum值'


def assert_series_unique(df: pd.DataFrame,
                         columns: (str, list) = '名称',
                         text: str = '',
                         ignore_na=False):
  """
  检查并输出重复项

  Args:
    df: 要检查的Dataframe
    columns: 唯一的列
    text: 输出的文案
    ignore_na: 是否忽略空值，默认False，为True时，有空值的行不参与校验
  """
  df = df.copy().reset_index(drop=True)
  columns = ensure_list(columns)
  if not is_unique_series(df, columns, ignore_na=ignore_na):
    if ignore_na:
      df = df[and_(*[df[c].notna() for c in columns])]
    df = df[df.duplicated(subset=columns)][columns].sort_values(columns)
    df = df.drop_duplicates()
    ls = df.astype(str).to_dict('records')
    ls = [','.join(list(i.values())) for i in ls]
    info = '\n->'.join(ls)
    raise AssertionError(f'{text}{columns}列中存在重复值:\n->{info}')


def assert_series_digit(df: pd.DataFrame, col: str):
  """检查一列是否可以转为数值型"""
  df = df.copy().reset_index(drop=True)
  values = df[col].unique().tolist()
  rv = [i for i in values if not is_digit(i) and not_empty(i)]
  assert not rv, f'"{col}"列应为数值型，{rv}无法转换为数值型'


def assert_series_not_like(df: pd.DataFrame, col: str, pattern):
  """检查并输出符合正则表达式pattern的值"""
  df = df.copy().reset_index(drop=True)
  values = df[col].unique().tolist()
  rv = [i for i in values if re.match(pattern, str(i)) and not_empty(i)]
  assert not rv, f'异常的"{col}": {rv}'


def assert_subset(values: (list, set, tuple), superset: (list, set, tuple)):
  """检查values是否是superset的子集"""
  res = [i for i in values if i not in superset]
  assert not res, f'{res}不在指定的集合中'



from ricco.util.strings import drop_repeat_string
from ricco.util.strings import get_breaks
from ricco.util.strings import get_single_list
from ricco.util.strings import is_repeated


def test_drop_repeat_string():
  assert drop_repeat_string(
      '112341234'
  ) == '11234'
  assert drop_repeat_string(
      '11'
  ) == '11'
  assert drop_repeat_string(
      '123123123123'
  ) == '123'
  assert drop_repeat_string(
      '上海市/普陀区/甘泉路街道/志丹路志丹路490号/志丹路志丹路490号'
  ) == '上海市/普陀区/甘泉路街道/志丹路490号'


def test_get_single_list():
  assert get_single_list(0, 7, 3) == [[0, 3], [3, 6], [6, 7]]


def test_get_breaks():
  assert get_breaks(10, 3) == [
    [[0, 3], [3, 6], [6, 9], [9, 10]],
    [[0, 1], [1, 4], [4, 7], [7, 10]],
    [[0, 2], [2, 5], [5, 8], [8, 10]]
  ]


def test_is_repeated():
  assert is_repeated('1111') is True
  assert is_repeated('1111', min_length=3) is False
  assert is_repeated('12121212') is True
  assert is_repeated('1212121') is False
  assert is_repeated('123123') is True
  assert is_repeated('123123', min_length=4) is False



import logging


def get_mp3_metadata(file_path):
  """获取MP3文件中的元数据信息"""

  def _get_v(dic, key):
    v = dic.get(key)
    if not v:
      return
    if isinstance(v, list):
      return '、'.join(v)
    return v

  from mutagen.easyid3 import EasyID3  # noqa

  try:
    audio = EasyID3(file_path)
  except Exception as e:
    logging.warning(f'Error: {e}')
    return {
      '标题': None, '艺术家': None, '专辑': None, '发行年份': None,
    }

  return {
    '标题': _get_v(audio, 'title'),
    '艺术家': _get_v(audio, 'artist'),
    '专辑': _get_v(audio, 'album'),
    '发行年份': _get_v(audio, 'date'),
  }



import sys


def ensure_osspath_valid(_path: str):
  if _path.startswith('oss://') and not _path.startswith('oss:///'):
    _path = _path.replace('oss://', 'oss:///')
  return _path


def ensure_dir(_path: str):
  if _path == '' or not _path:
    return ''
  if _path.endswith('/'):
    return _path
  return f'{_path}/'


def bucket_from_path(_path):
  _path = ensure_osspath_valid(_path)
  _path = _path.lstrip('oss:///')
  return _path.split('/')[0]


def progress_bar(consumed, total):
  """进度条"""
  if total:
    rate = str(round(100 * (float(consumed) / float(total)), 2))
    print(f'\r传输中: {rate}% ', end='')
    sys.stdout.flush()



from ricco.util.os import ensure_dir
from ricco.util.os import extension
from ricco.util.os import path_name
from ricco.util.os import split_path


def test_path_name():
  assert path_name('test_util.py') == 'test_util'


def test_extension():
  assert extension('test_util.py') == '.py'


def test_split_path():
  assert split_path('/path/test.txt') == ('/path', 'test', '.txt')
  assert split_path('/path/2/test.txt') == ('/path/2', 'test', '.txt')


def test_ensure_dir():
  assert ensure_dir('/path/') == '/path/'
  assert ensure_dir('/path') == '/path/'



import os.path

from ..base import ensure_list
from ..util.os import extension
from ..util.os import split_path
from .util import bucket_from_path
from .util import ensure_dir
from .util import ensure_osspath_valid
from .util import progress_bar


def oss_client(
    *,
    bucket_name,
    access_key,
    secret_key,
    endpoint=None,
    region=None
):
  """
  创建oss client

  Args:
    bucket_name: bucket名称
    access_key: access_key
    secret_key: secret_key
    endpoint: oss endpoint，默认为“https://oss-cn-hangzhou.aliyuncs.com”
    region: oss region，默认为“cn-hangzhou”
  """
  import oss2
  return oss2.Bucket(
      auth=oss2.Auth(access_key, secret_key),
      endpoint=endpoint or 'https://oss-cn-hangzhou.aliyuncs.com',
      bucket_name=bucket_name,
      region=region or 'cn-hangzhou'
  )


class OssUtils:
  def __init__(self, *, work_path, access_key, secret_key,
               endpoint=None, region=None):
    """
    oss工具类，所有操作均可使用完整路径或work_path的相对路径

    Args:
      work_path: 工作目录，以“oss:///”开头
      access_key: 有work_path权限的access_key
      secret_key: 有work_path权限的secret_key
      endpoint: oss endpoint，默认为“https://oss-cn-hangzhou.aliyuncs.com”
      region: oss region，默认为“cn-hangzhou”
    """
    import oss2
    self.oss2 = oss2
    self.work_path = ensure_dir(ensure_osspath_valid(work_path))
    self.bucket_name = bucket_from_path(work_path)
    self.client = oss_client(
        bucket_name=self.bucket_name,
        access_key=access_key,
        secret_key=secret_key,
        endpoint=endpoint,
        region=region
    )

  @property
  def _prefix(self):
    """oss前缀 + bucket"""
    return f'oss:///{self.bucket_name}/'

  def abspath(self, _path):
    if _path.startswith('oss://'):
      _path = ensure_osspath_valid(_path)
      assert _path.startswith(
          self.work_path), f'必须为“{self.work_path}”的子目录'
      return _path
    return f'{self.work_path}{_path}'

  def object_path(self, _path):
    """将文件路径转为oss2的输入路径，即Bucket之后的路径"""
    _path = self.abspath(_path)
    if _path.endswith('/'):
      return os.path.relpath(_path, self._prefix) + '/'
    return os.path.relpath(_path, self._prefix)

  def upload(self, path_local, path_remote=None, *, overwrite=False):
    """
    上传文件到oss中

    Args:
      path_local: 本地文件路径
      path_remote: oss路径及文件名，若不指定则上传为work_path下的同名文件
    """
    if not path_remote:
      path_remote = ''.join(split_path(path_local)[-2:])
    path_remote = self.object_path(path_remote)
    if not overwrite and self.exist(self._prefix + path_remote):
      raise FileExistsError(f'文件已存在：{path_remote}')
    self.client.put_object_from_file(
        path_remote,
        path_local,
        progress_callback=progress_bar,
    )

  def download(self, path_remote, path_local=None, *, overwrite=False):
    """
    从oss中下载文件

    Args:
      path_remote: oss路径及文件名
      path_local: 本地文件路径，若不指定则下载为work_path下的同名文件
    """
    if not path_local:
      path_local = ''.join(split_path(path_remote)[-2:])
    if not overwrite and os.path.exists(path_local):
      raise FileExistsError(f'文件已存在：{path_remote}')
    path_remote = self.object_path(path_remote)
    self.client.get_object_to_file(
        path_remote, path_local,
        progress_callback=progress_bar,
    )

  def mkdir(self, dir_path):
    """创建目录"""
    dir_path = ensure_dir(dir_path)
    dir_path = self.object_path(dir_path)
    self.client.put_object(dir_path, '')

  def rm(self, _path):
    """删除文件或目录"""
    self.client.delete_object(self.object_path(_path))

  def exist(self, _path):
    """判断文件是否存在"""
    return self.client.object_exists(self.object_path(_path))

  def dir_iter(self, dir_path=None, exts=None, recursive=False):
    """
    遍历目录下的文件

    Args:
      dir_path: 目录路径，如不指定则遍历实例化时指定的work_path
      exts: 文件后缀
      recursive: 是否递归遍历
    """
    exts = ensure_list(exts)
    dir_path = ensure_dir(dir_path)
    dir_path = self.object_path(dir_path)
    for obj in self.oss2.ObjectIterator(
        self.client, prefix=dir_path,
        delimiter='' if recursive else '/'
    ):
      if obj.is_prefix():
        continue
      if exts and extension(obj.key) not in exts:
        continue
      if obj.key.endswith('/'):
        continue
      yield self._prefix + obj.key



from functools import lru_cache

from ..base import is_empty
from ..base import warn_
from ..resource.postcode import get_postcode
from .district import norm_city_name


@lru_cache()
def postcode_city_mapping():
  """
  获取邮编与城市名称的映射关系
  """
  df = get_postcode()[['postcode', 'city']]
  df = df.dropna()
  df = df.drop_duplicates(subset=['postcode'])
  df = df.set_index('postcode')
  return df.to_dict()['city']


@lru_cache()
def get_city_from_postcode(postcode: str, full=False, warning=False):
  """
  根据邮编获取城市名称

  Args:
    postcode: 邮编
    full: 是否返回全称
    warning: 是否打印警告
  """
  if is_empty(postcode):
    return
  if isinstance(postcode, (float, int)):
    postcode = str(int(postcode))
    postcode = postcode.zfill(6)
  if not isinstance(postcode, str) or len(postcode) != 6:
    warn_(f'邮编{postcode}必须为字符串且长度为6', if_or_not=warning)
    return
  mapping = postcode_city_mapping()
  city = mapping.get(postcode)
  if city:
    return norm_city_name(city, full=full)



import builtins
import json
import logging
import re
import uuid
import warnings
from ast import literal_eval
from datetime import datetime
from itertools import chain
from itertools import groupby
from typing import Iterable

import numpy as np
import pandas as pd
from fuzzywuzzy import fuzz

from ..base import ensure_list
from ..base import is_empty
from ..base import not_empty
from .decorator import check_null
from .decorator import check_str


def to_json_string(string, errors='raise'):
  """将字符串转为json格式的字符串"""
  assert errors in ('raise', 'coerce', 'ignore'), '参数错误'
  if not isinstance(string, (list, dict)):
    try:
      string = literal_eval(string)
    except ValueError as e:
      if errors == 'raise':
        raise ValueError(e)
      if errors == 'coerce':
        return
      if errors == 'ignore':
        return string

  return json.dumps(string, ensure_ascii=False)


@check_str
def relstrip(string, kwd):
  """通过正则表达式删除左侧字符串"""
  return re.sub(f'^{kwd}', '', string)


@check_str
def rerstrip(string, kwd):
  """通过正则表达式删除右侧字符串"""
  return re.sub(f'{kwd}$', '', string)


def get_shortest_element(elements: list):
  """获取列表中长度最短的元素"""

  @check_null(default_rv=np.inf)
  def condition(string):
    string = str(string)
    return len(string)

  return min(elements, key=condition)


def and_(*conditions):
  """对多个条件执行and操作"""
  assert len(conditions) >= 1, '最少一个条件'
  res = conditions[0]
  for c in conditions[1:]:
    res = res & c
  return res


def or_(*conditions):
  """对多个条件执行or操作"""
  cond = []
  for c in conditions:
    if isinstance(c, (list, tuple)):
      cond.extend(list(c))
    else:
      cond.append(c)
  assert len(cond) >= 1, '最少一个条件'
  res = cond[0]
  for c in cond[1:]:
    res = res | c
  return res


def physical_age(birthday: datetime, deadline: datetime = None):
  """计算周岁，默认按照当前时间点计算"""
  now = deadline or datetime.now()
  now_str = now.strftime('%m%d')
  birth_str = birthday.strftime('%m%d')
  res = now.year - birthday.year
  if now_str < birth_str:
    res -= 1
  return res


def first_notnull_value(series: Iterable):
  """筛选第一个不为空的值"""
  for v in series:
    if not_empty(v):
      return v
  warnings.warn('所有值均为空值')


@check_null()
def pinyin(word: str) -> str:
  """将中文转换为汉语拼音"""
  import pypinyin
  assert isinstance(word, str), '输入参数必须为字符串'
  s = ''
  for i in pypinyin.pinyin(word, style=pypinyin.NORMAL):
    s += ''.join(i)
  return s


@check_null(default_rv=False)
def is_valid_uuid(uuid_to_test, version=4):
  """判断一个字符是否是uuid"""
  try:
    uuid_obj = uuid.UUID(uuid_to_test, version=version)
  except ValueError:
    return False
  return str(uuid_obj) == uuid_to_test


@check_null(default_rv=uuid.uuid4())
def get_uuid(s):
  """针对格式错误的uuid和空白值生成新的uuid"""
  if is_valid_uuid(s):
    return s
  return uuid.uuid4()


@check_null(default_rv=np.nan)
def per2float(string: str) -> float:
  """带有百分号的数值字符串转小数点形式的数值，没有百分号的返回原值"""
  if string.endswith('%'):
    string = string.rstrip('%')
    return float(string) / 100
  return float(string)


@check_null()
def extract_num(string: str,
                num_type: str = 'str',
                method: str = 'list',
                join_list: bool = False,  # noqa
                ignore_pct: bool = True,
                multi_warning=False):
  """
  提取字符串中的数值，默认返回所有数字组成的列表

  Args:
    string: 输入的字符串
    num_type:  输出的数字类型，int/float/str，默认为str
    method: 结果计算方法，对结果列表求最大/最小/平均/和等，numpy方法，默认返回列表本身
    join_list: 是否合并列表，默认FALSE
    ignore_pct: 是否忽略百分号，默认True
    multi_warning: 当有多个值的时候是否输出警告信息
  """
  assert num_type in ('float', 'int', 'str')
  string = str(string)
  ls = re.findall(r'\d+\.?\d*' if ignore_pct else r'\d+\.?\d*%?', string)
  if not ls:
    return
  if method == 'list':
    return ls if num_type == 'str' else [
      getattr(builtins, num_type)(per2float(i)) for i in ls
    ]
  if multi_warning and len(ls) >= 2:
    warnings.warn(f'有多个值进行了"{method}"运算')
  ls = [float(per2float(i)) for i in ls]
  return getattr(np, method)(ls)


@check_null()
def to_float(string,
             rex_method: str = 'mean',
             ignore_pct: bool = False,
             multi_warning=True):
  """字符串转换为float"""
  return extract_num(string,
                     num_type='float',
                     method=rex_method,
                     ignore_pct=ignore_pct,
                     multi_warning=multi_warning)


@check_null()
def house_type_format(x):
  """
  通过正则表达式将户型统一修改为1房，2房···5房及以上，目前只支持9室以下户型，
  其中5室及以上的类别为“5房及以上”
  """
  from ..resource import UTIL_CN_NUM
  if re.match('^[1-4][室房]$', x) or x == '5房及以上':
    return x
  exp = ''.join(UTIL_CN_NUM.keys())
  if res_num := re_fast(fr'([{exp}\d+])[室房]', str(x)):
    for ori, dst in UTIL_CN_NUM.items():
      res_num = res_num.replace(ori, dst)
    if int(res_num) <= 4:
      return f'{res_num}房'
    else:
      return '5房及以上'


@check_null()
def segment(x: (int, float),
            gap: (list, tuple, set, float, int),
            sep: str = '-',
            unit: str = '',
            bottom: str = '以下',
            top: str = '以上') -> str:
  """
  区间段划分工具

  Args:
    x: 数值
    gap: 间隔，固定间隔或列表
    unit: 单位，末尾
    sep: 分隔符，中间
    bottom: 默认为“以下”：80米以下
    top: 默认为“以上”：100米以上
  Returns:
    区间段 'num1分隔符num2单位'：‘80-100米’
  """

  def between_list(_x, lis):
    for i in reversed(range(len(lis) - 1)):
      if _x >= lis[i]:
        return lis[i], lis[i + 1]

  assert isinstance(gap, (list, tuple, set, int, float))

  x = to_float(x)
  if isinstance(gap, (list, tuple, set)):
    gap = sorted(list(set(gap)))
    if x < gap[0]:
      return f'{gap[0]}{unit}{bottom}'
    if x >= gap[-1]:
      return f'{gap[-1]}{unit}{top}'
    return f'{between_list(x, gap)[0]}{sep}{between_list(x, gap)[1]}{unit}'

  if isinstance(gap, (int, float)):
    if x >= 0:
      return f'{int(x / gap) * gap}{sep}{int(x / gap) * gap + gap}{unit}'
    else:
      return f'{int(x / gap) * gap - gap}{sep}{int(x / gap) * gap}{unit}'


def to_str_list(series: (list, pd.Series, tuple)) -> list:
  """将列表中的元素保留为字符串、唯一、非空"""
  return list(set([str(i) for i in series if not_empty(i)]))


@check_null(default_rv=(None, None, None, None))
def fuzz_match(string: str,
               string_set: (list, pd.Series, tuple),
               fix_string_set: bool = False,
               valid_score: int = 0):
  """
  为某一字符串从某一集合中匹配相似度最高的元素

  Args:
    string: 输入的字符串
    string_set: 要去匹配的集合
    fix_string_set: 是否修复string_set中的异常数据，使用该选项会降低性能
    valid_score: 相似度大于该值的才返回
  Returns: 字符串及相似度组成的列表
  """
  w = 0.3

  def score(x, s):
    return fuzz.partial_ratio(x, s) + fuzz.ratio(x, s) * w

  if fix_string_set:
    string_set = to_str_list(string_set)
  if is_empty(string_set):
    return None, None, None, None
  if string in string_set:
    return string, 100, 100, 100.0
  max_s = max(string_set, key=lambda x: score(x, string))

  ratio = fuzz.ratio(string, max_s)
  p_ratio = fuzz.partial_ratio(string, max_s)
  if ratio < valid_score and p_ratio < valid_score:
    return None, None, None, None
  return max_s, ratio, p_ratio, round((ratio * w + p_ratio) / (1 + w), 2)


def get_city_id_by_name(city: str):
  """获取城市代码"""
  from ..resource.city_id import CITY_IDS
  city_id = CITY_IDS.get(city if city.endswith('市') else f'{city}市')
  if not city_id:
    logging.warning('获取city_id失败，请补充该城市')
  return city_id


def sort_by_list(src_list, by_list, filter_=False) -> list:
  """
  根据一个列表对另一个列表进行筛选或排序，参照列表中不存在的元素按照原始顺序排列在后

  Args:
    src_list: 要进行排序的列表
    by_list: 参照的列表
    filter_: 是否根据参照列表筛选

  Examples:
    >>> a = [1, 2, 3, 4, 5]
    >>> b = [2, 5, 4, 1]
    >>> sort_by_list(a, b)
    [2, 5, 4, 1, 3]
    >>> sort_by_list(a, b, filter_=True)
    [2, 5, 4, 1]
    >>> sort_by_list(b, a)
    [1, 2, 4, 5]
  """
  res = [i for i in by_list if i in src_list]
  if not filter_:
    _res = [i for i in src_list if i not in res]
    res.extend(_res)
  return res


def remove_null_in_dict(dic: dict) -> dict:
  """将字典中值为空的元素删掉"""
  return {
    k: v
    for k, v in dic.items()
    if not_empty(v)
  }


def union_str_v2(*strings, sep='') -> str:
  """
  连接字符串，空白字符串会被忽略

  Args:
    *strings: 要连接字符串，依次传入
    sep: 连接符，默认为空白字符串

  Examples:
    >>> union_str_v2('a', 'b', sep='-')
    'a-b'
    >>> union_str_v2('a', 'b', 'c')
    'abc'

  """
  if strings := [i for i in strings if not_empty(i) and i != '']:
    return sep.join(strings)


def union_list_v2(*lists) -> list:
  """
  合并列表

  Examples:
    >>> a = [1]
    >>> b = [2, 3]
    >>> union_list_v2(a, b)
    [1, 2, 3]
  """
  lists = [ensure_list(i) for i in lists if not_empty(i)]
  return list(chain(*lists))


@check_null()
def eval_(x: str):
  """将文本型的列表、字典等转为真正的类型"""
  return literal_eval(x)


@check_null(default_rv={})
def list2dict(x: list):
  """列表转为字典，key为元素的顺序"""
  return {i: j for i, j in enumerate(x)}


def re_fast(pattern, string, warning=True):
  """根据正则表达式快速提取匹配到的第一个"""
  if isinstance(string, str):
    if ls := re.findall(pattern, string):
      if warning and len(ls) >= 2:
        warnings.warn('匹配到多个值，默认返回第一个')
      return ls[0]


@check_null()
def rstrip_d0(x):
  """删除末尾的‘.0’,并转为str格式，适用于对手机号等场景，如：'130.0' -> '130'"""
  _x = str(x)
  if re.match(r'^\d+\.0$', _x):
    return _x[:-2]
  return x


@check_null()
def fix_str(x: str) -> (str, None):
  """将字符串两端的空格及换行符删除，如果为空白字符串则返回空值"""
  if isinstance(x, str):
    x = x.strip()
    if x == '':
      return
  return x


def interchange_dict(dic: dict) -> dict:
  """将字典的key和value互换"""
  return dict((v, k) for k, v in dic.items())


def to_bool(x,
            na=False,
            other=False,
            t_list: list = None,
            f_list: list = None):
  """
  将常见的布尔类型的代替值转为布尔类型

  Args:
    x: 输入值
    na: 空值返回真还是假
    other: 无法判断的值如何处理
      - 'raise'：抛出异常
      - 'coerce'：返回传入的值
      - 除 'raise' 和 'coerce' 之外的其他值：直接返回该值
    t_list:指定为True的类别
    f_list:指定为False的类别
  """
  if not t_list:
    t_list = ['是', 1, 1.0, '1', '1.0', 't', 'true']
  if not f_list:
    f_list = ['否', 0, '0', 'f', 'false']

  if is_empty(x):
    return na
  if isinstance(x, bool):
    return x

  x2 = x.lower() if isinstance(x, str) else x
  if x2 in t_list:
    return True
  if x2 in f_list:
    return False
  if other == 'raise':
    raise ValueError(f'无法转换的值：{x}')
  if other == 'coerce':
    return x
  return other


def is_unique_series(df: pd.DataFrame,
                     key_cols: (str, list) = None,
                     ignore_na=False):
  """判断Dataframe中的某一列或某几列的组合是否唯一"""
  df = df.copy()
  if not key_cols:
    key_cols = df.columns.tolist()
  key_cols = ensure_list(key_cols)
  if ignore_na:
    df = df[and_(*[df[c].notna() for c in key_cols])]
  return not df.duplicated(subset=key_cols, keep=False).any()


@check_null(default_rv=False)
def is_digit(x) -> bool:
  """判断一个值是否能通过float方法转为数值型"""
  if isinstance(x, (float, int)):
    return True
  if isinstance(x, str):
    if re.match(r'\d+\.?\d*', x):
      return True
  return False


@check_null(default_rv=False)
def is_hex(string) -> bool:
  """判断一个字符串是否是十六进制格式"""
  if re.match('^[0-9a-fA-F]+$', str(string)):
    return True
  return False


def isinstance_in_list(values: list, types: (str, list)) -> bool:
  """
  检查列表内的元素类型是否全部满足多个类型之一

  Args:
    values: 要检查的值
    types: 类型类别
  """
  values = ensure_list(values)
  return all([isinstance(v, types) for v in values])


def drop_repeat(x: (list, tuple)):
  """
  删除列表中连续重复的元素

  Examples:
    >>> drop_repeat([1, 2, 2, 3, 4, 4, 4, 3])
    [1, 2, 3, 4, 3]
  """
  return [key for key, _ in groupby(x)]


def diff(ls1: list, ls2: list):
  """求两个列表的差集"""
  return list(set(ls1) - set(ls2)), list(set(ls2) - set(ls1))


def fuzz_pair(ls1, ls2, score=50):
  """模糊配对"""
  mark = 0 if len(ls1) <= len(ls2) else 1
  ls_iter = ls1 if mark == 0 else ls2
  ls_set = ls1 if mark != 0 else ls2

  res_list = []
  for i in ls_iter:
    rv = fuzz_match(i, ls_set)
    if rv[3] is not None and rv[3] > score:
      res_list.append([i, rv[0]] if mark == 0 else [rv[0], i])
  return res_list


def check_diff(ls1: list, ls2: list):
  """检查两个列表的差集"""
  res1, res2 = diff(ls1, ls2)
  print(f'左侧特有元素：{res1}')
  print(f'\n右侧特有元素：{res2}')
  print(f'\n模糊配对结果：{fuzz_pair(res1, res2)}')


def fix_sql(s):
  """移除SQL语句中的换行符和多余的空格"""
  s = s.replace('\t', ' ').replace('\n', ' ')
  while '  ' in s:
    s = s.replace('  ', ' ')
  return s.strip()



import os.path
import stat

from ..base import ensure_list
from ..util.os import ensure_dirpath_exist
from ..util.os import extension


def get_sftp_client(
    *,
    hostname,
    port,
    username,
    password
):
  import paramiko
  ssh = paramiko.SSHClient()
  ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
  ssh.connect(hostname=hostname,
              port=port,
              username=username,
              password=password)
  return ssh.open_sftp()


class SFTP:
  def __init__(self, *, hostname, port, username, password):
    self._sftp = get_sftp_client(
        hostname=hostname,
        port=port,
        username=username,
        password=password
    )

  @property
  def workdir(self):
    return self._sftp.getcwd() or '/'

  def isfile(self, path):
    """判断是否为文件"""
    return stat.S_ISREG(self._sftp.stat(path).st_mode)

  def isdir(self, path):
    """判断是否为目录"""
    return stat.S_ISDIR(self._sftp.stat(path).st_mode)

  def chdir(self, path):
    """切换工作目录"""
    self._sftp.chdir(path)
    print(self.listdir())

  def listdir(self, path=None, recursive=False):
    """列出目录下的所有文件和子目录，返回列表"""
    path = path or self.workdir
    if not recursive:
      return self._sftp.listdir(path)
    else:
      dirs = []
      for item in self._sftp.listdir(path):
        item_path = os.path.join(path, item)
        if self.isdir(item_path):
          dirs.append(item_path)
          dirs.extend(self.listdir(item_path, recursive=True))
        else:
          dirs.append(item_path)
      return sorted(dirs)

  def dir_iter(self, path=None, recursive=False, exts=None):
    """遍历目录，返回所有文件路径 """
    for item in self.listdir(path, recursive):
      if self.isfile(item):
        if exts:
          exts = ensure_list(exts)
          if extension(item) in exts:
            yield item
        else:
          yield item

  def download(self, remote_path, local_dir):
    """下载单个文件或整个目录"""
    assert os.path.isdir(local_dir), '本地路径必须是一个目录'
    os.makedirs(local_dir, exist_ok=True)
    if self.isfile(remote_path):
      filename = os.path.basename(remote_path)
      local_path = os.path.join(local_dir, filename)
      ensure_dirpath_exist(local_path)
      self._sftp.get(remote_path, local_path)
    elif self.isdir(remote_path):
      files = [
        i for i in self.listdir(remote_path, recursive=True) if self.isfile(i)
      ]
      print(f'共{len(files)}个文件')
      for i, file in enumerate(files):
        print(f'正在下载第{i + 1}个文件，{file}')
        objname = os.path.basename(remote_path.rstrip('/'))
        subpath = os.path.relpath(file, remote_path)
        local_path = os.path.join(local_dir, objname, subpath)
        ensure_dirpath_exist(local_path)
        self._sftp.get(file, local_path)
    else:
      raise Exception(f'未知的路径类型：{remote_path}')



"""资源目录"""
import os


def resource_path(filename):
  return os.path.join(__path__[0], filename)


P_BD_REGION = resource_path('bd_region.csv')
P_POSTCODE = resource_path('postcode.csv')
P_STOPWORDS = resource_path('stopwords.txt')

UTIL_CN_NUM = {
  '〇': '0',
  '零': '0',
  '一': '1',
  '二': '2',
  '两': '2',
  '三': '3',
  '四': '4',
  '五': '5',
  '六': '6',
  '七': '7',
  '八': '8',
  '九': '9',
}



import csv
import os
import shutil

import fiona
import pandas as pd
from shapely.geometry import shape
from tqdm import tqdm

from ..base import ensure_ext
from ..base import warn_
from ..geometry.util import dumps2x
from ..util.os import dir_iter_list
from ..util.os import ensure_dir
from ..util.os import ensure_dirpath_exist
from ..util.os import extension
from ..util.os import path_name
from ..util.os import remove_path
from ..util.os import single_ext
from ..util.os import split_path
from .extract import rdf
from .extract import rdf_by_dir
from .load import to_file
from .load import to_parts_file


def split2x_by_parts(
    filename: str,
    *,
    parts: int = None,
    to_ext='.csv',
    log=True,
    fn=None,
):
  """
  将文件拆分为多个文件，放置在与文件同名目录下

  Args:
    filename: 文件路径
    parts: 输出的文件个数
    to_ext: 输出文件扩展名
    log: 是否输出文件保存信息
    fn: 处理函数
  """
  # 创建目标目录
  to_ext = ensure_ext(to_ext)
  full_path = os.path.abspath(filename)
  dirpath = path_name(full_path)
  # 读取文件并分批保存
  df = rdf(full_path)
  if fn:
    df = fn(df)
  to_parts_file(df, dirpath, parts=parts, to_ext=to_ext, log=log)


def split2x_by_chunksize(
    filepath: str,
    *,
    chunksize: int = 50000,
    to_ext='.csv',
    log=True,
    fn=None,
    **kwargs,
):
  """
  将文件拆分为多个文件，放置在与文件同名目录下

  Args:
    filepath: 输入的文件路径
    chunksize: 每份文件的数据量
    to_ext: 输出文件扩展名
    log: 是否输出文件保存信息
    fn: 处理函数
  """
  to_ext = ensure_ext(to_ext)
  dirpath = path_name(os.path.abspath(filepath))
  ext = extension(filepath)
  if ext == '.csv':
    for i, _df in enumerate(
        pd.read_csv(filepath, chunksize=chunksize, dtype=str)
    ):
      if fn:
        _df = fn(_df)
      savefile = os.path.join(dirpath, f'part_{str(i).zfill(6)}{to_ext}')
      to_file(_df, savefile, log=log, **kwargs)
  else:
    df = rdf(filepath)
    if fn:
      df = fn(df)
    to_parts_file(df, dirpath,
                  chunksize=chunksize, to_ext=to_ext,
                  log=log)


def split2x(
    filename: str,
    *,
    chunksize: int = None,
    parts: int = None,
    to_ext='.csv',
    log=True,
    fn=None,
):
  """
  将文件拆分为多个文件，放置在与文件同名目录下

  Args:
    filename: 输入的文件路径
    chunksize: 每份文件的数据量
    parts: 输出的文件个数
    to_ext: 输出的文件扩展名
    log: 是否输出文件保存信息
    fn: 处理函数
  """
  assert any([chunksize, parts]), 'chunksize 和 parts必须指定一个'
  to_ext = ensure_ext(to_ext)
  if chunksize:
    split2x_by_chunksize(filename, chunksize=chunksize, to_ext=to_ext, log=log,
                         fn=fn)
  if parts:
    split2x_by_parts(filename, parts=parts, to_ext=to_ext, log=log, fn=fn)


def file_to_x(filepath, to_ext,
              to_dir=None,
              delete=False,
              overwrite=False,
              log=True,
              **kwargs):
  """
  文件格式转换，整体读取并转换，将文件整体加载到内存中再保存为另一种文件格式

  Args:
    filepath: 待转换的文件路径
    to_ext: 要保存的文件扩展名
    to_dir: 要写入的文件路径，默认存放在原目录
    delete: 是否删除原文件，默认不删除
    overwrite: 是否覆盖原文件，默认不覆盖
    log: 是否输出文件保存、删除日志
  """
  to_ext = ensure_ext(to_ext)
  dirpath, filename, ex = split_path(filepath)
  if not to_dir:
    to_dir = dirpath
  to_file_path = os.path.join(to_dir, f'{filename}{to_ext}')
  if not overwrite and os.path.exists(to_file_path):
    warn_(f'文件{filepath}已存在，已跳过', log, 'logging')
    return
  if ex == to_ext:
    ensure_dirpath_exist(to_file_path)
    shutil.copyfile(filepath, to_file_path)
  to_file(rdf(filepath, dtype=str), to_file_path, log=log, **kwargs)
  if delete:
    os.remove(filepath)
    warn_(f'Deleted: {filepath}', log, 'logging')


def dir_file_to_x(from_dir, to_dir,
                  from_ext=None, to_ext=None,
                  recursive=False,
                  delete=False,
                  overwrite=False,
                  log=False):
  """转换整个目录中的文件格式"""
  assert from_dir != to_dir, 'from_dir and to_dir must be different'
  assert to_ext is not None, 'to_ext must be specified'
  assert from_ext != to_ext, 'from_ext and to_dir must be different'
  to_ext, from_ext = ensure_ext(to_ext), ensure_ext(from_ext)
  path_list = dir_iter_list(from_dir, exts=from_ext, recursive=recursive)
  print(len(path_list))
  for f in tqdm(path_list):
    file_to_x(
        f, to_ext, to_dir=to_dir, delete=delete,
        overwrite=overwrite, log=log,
    )


def reshape_files(from_dir, to_dir,
                  from_ext=None, to_ext=None,
                  chunksize: int = 100000,
                  func=None,
                  log=False,
                  **kwargs):
  """
  将文件拆分成小文件，并保存到to_dir中

  Args:
    from_dir: 读取的目录
    to_dir: 保存的目录
    from_ext: 读取的扩展名，默认为None，全部读取
    to_ext: 保存的扩展名，默认为None，若读取文件的扩展名唯一，则使用读文文件的扩展名，反之需要指定
    chunksize: 每个文件的大小
    func: 自定义处理函数
    log: 是否打印日志
  """
  from_dir, to_dir = ensure_dir(from_dir), ensure_dir(to_dir)
  assert from_dir != to_dir, 'from_dir and to_dir must be different'
  assert chunksize >= 1
  n, total_lines, after_lines = 1, 0, 0
  df = pd.DataFrame()
  path_list = dir_iter_list(from_dir, exts=from_ext)
  # 扩展名在不指定输出的情况下，从输入的扩展名中获取
  if not to_ext:
    if ext := single_ext(path_list):
      to_ext = ext
    else:
      raise ValueError('文件中有多个不同扩展名的文件，请指定to_ext')
  for p in tqdm(path_list):
    df_temp = rdf(p)
    total_lines += df_temp.shape[0]
    if func:
      df_temp = func(df_temp)
    after_lines += df_temp.shape[0]
    df = pd.concat([df, df_temp], ignore_index=True)
    while df.shape[0] > chunksize:
      to_file(df[df.index < chunksize],
              os.path.join(to_dir, f'part_{str(n).zfill(6)}{to_ext}'),
              log=log, **kwargs)
      df = df[df.index >= chunksize].reset_index(drop=True)
      n += 1
  to_file(df, os.path.join(to_dir, f'part_{str(n).zfill(6)}{to_ext}'),
          log=log,
          **kwargs)
  print(f'输入文件总数: {len(path_list)}, 输入数据量: {total_lines}')
  print(f'输出文件总数: {n}, 输出数据量: {after_lines}')


def df_iter_by_column(df: pd.DataFrame, by, na='null'):
  """按列的值分组迭代df"""
  for enum, _df in df.groupby(by, as_index=False, dropna=False):
    yield enum if pd.notna(enum) else na, _df


def split_csv_by_column(
    filepath: str, by: str,
    *,
    to_dir: str = None,
    to_ext: str = '.csv',
    merge_file: bool = True,
    chunksize=100000,
):
  """
  将csv文件按指定列的值拆分为多个文件

  Args:
    filepath: csv文件路径
    by: 根据那一列拆分
    to_dir: 要保存的位置，默认与原文件同名目录
    to_ext: 输出的文件扩展名
    merge_file: 是否将每一类数据合并为一个文件，默认合并
    chunksize: 中间文件大小，如内存不足则可适当调低
  """
  dirpath, filename, ex = split_path(filepath)
  to_dir = to_dir or f'{dirpath}/{filename}'
  # 作为中间文件时使用parquet提高速度
  part_ext = '.parquet' if merge_file else to_ext
  print(f'根据"{by}"列将大文件拆分至对应文件夹')
  for i, df in enumerate(pd.read_csv(filepath, chunksize=chunksize, dtype=str)):
    for name, _df in df_iter_by_column(df, by):
      to_file(_df, f'{dirpath}/{filename}/{name}/{i}{part_ext}', log=False)
  if merge_file:
    print('将各个文件夹中的数据合并为一个文件')
    for name in os.listdir(to_dir):
      to_file(
          rdf_by_dir(f'{to_dir}/{name}', exts='.parquet'),
          f'{to_dir}/{name}{to_ext}',
      )
      remove_path(f'{to_dir}/{name}')


def merge_csv_files(dir_path, output_file):
  """
  合并指定目录下的所有CSV文件到一个输出文件中，极低内存消耗

  参数:
  - csv_dir: CSV文件所在的目录
  - output_file: 合并后的输出CSV文件路径
  """
  path_list = dir_iter_list(dir_path, exts='.csv')

  if not path_list:
    print("没有找到csv文件，请检查目录。")

  # 将列表中的第一个csv文件复制并重命名为输出文件
  shutil.copyfile(path_list[0], output_file)

  # 从列表的第二个CSV文件开始逐个处理
  for p in path_list[1:]:
    print(f"正在合并 {p} 到 {output_file} ...")
    with open(p, 'r') as fin:
      # 跳过表头
      next(fin)
      # 以追加模式打开输出文件，并逐行写入内容
      with open(output_file, 'a') as fout:
        for line in fin:
          fout.write(line)
  print(f"所有csv文件已经合并为 '{output_file}'。")


def gdb2csv(
    dir_path,
    output_path=None,
    with_geometry=True,
    geom_format='wkb',
    log=True):
  """gdb文件转换为csv文件"""
  if not output_path:
    output_path = f'{dir_path}.csv'
  if log:
    print(f'Saving: {output_path}')
  with fiona.open(dir_path) as src:
    # 获取图层的字段名称
    columns = list(src.schema['properties'].keys())
    if with_geometry:
      columns.append('geometry')
    with open(output_path, mode='w', newline='', encoding='utf-8') as csv_file:
      writer = csv.DictWriter(csv_file, fieldnames=columns)
      writer.writeheader()
      # 逐行读取GDB文件中的记录
      for line in tqdm(src):
        r = {
          c: line['properties'][c] for c in columns if c != 'geometry'
        }
        if with_geometry:
          r['geometry'] = dumps2x(shape(line['geometry']), geom_format)
        writer.writerow(r)



from functools import lru_cache

import pandas as pd

from . import P_STOPWORDS


@lru_cache()
def get_stopwords() -> list:
  """获取停词表列表"""
  print('获取停词表')
  df = pd.read_csv(P_STOPWORDS, header=None)
  return df[0].tolist()



import warnings
from itertools import combinations

import geopandas as gpd
import pandas as pd
from shapely.errors import ShapelyDeprecationWarning
from shapely.geometry import MultiPolygon
from tqdm import tqdm

from ..util.assertion import assert_not_null
from ..util.assertion import assert_subset
from .df import auto2shapely
from .util import ensure_valid_polygon
from .util import filter_polygon_from_collection

warnings.filterwarnings('ignore', category=ShapelyDeprecationWarning)

_desc = '''
  拓扑检查
  可检查: 重叠问题, 自相交问题, geometry缺失问题
  可修复: 重叠问题, 自相交问题
'''


def is_topology_valid(df: gpd.GeoDataFrame, c_geometry='geometry'):
  """
  判断整列geometry是否规范, 仅支持面数据，
    * 检查geometry是否有空值，
    * 检查每个geometry是否存在拓扑问题，
    * 检查整列geometry之间是否存在拓扑问题

  Args:
    df: 需要进行拓扑修复的GeoDataFrame
    c_geometry: geometry列名

  Returns:
    geometry数据是否规范
  """
  df = auto2shapely(df, geometry=c_geometry)
  assert_not_null(df, c_geometry)
  assert_subset(df.geom_type.unique(), {'Polygon', 'MultiPolygon'})

  if not all(df.is_valid):
    return False

  n = df.shape[0]
  for x, y in tqdm(combinations(df.geometry, 2), total=n * (n - 1) / 2,
                   desc='topology check'):
    if x.intersects(y):
      return False
  return True


def _fix_topology(series: pd.Series,
                  fill_intersects=True,
                  keep_contains=True):
  """
  Args:
    series: series的index不能重复, 如果keep_contains=True,geometry不能相互包含
    fill_intersects: 是否填满相交的区域，如果为True的话，两个面相交的区域会分配到更靠后的面内
    keep_contains: fill_intersects为True时才有效，即是否强制保留被包含的面
  """

  def ser_intersect(geo_ser: pd.Series):
    intersect_info = {index: {'intersect_list': [], 'is_contains': False}
                      for index in geo_ser.index}
    size = geo_ser.shape[0]
    for index_x, index_y in tqdm(combinations(geo_ser.index, 2),
                                 total=size * (size - 1) / 2,
                                 desc='intersect info'):
      geo_x = geo_ser.loc[index_x]
      geo_y = geo_ser.loc[index_y]
      if geo_x.intersects(geo_y):
        intersect_info[index_x]['intersect_list'].append(index_y)
        intersect_info[index_y]['intersect_list'].append(index_x)

    for index, info in tqdm(intersect_info.items(),
                            total=len(intersect_info.keys()),
                            desc='contains info'):
      intersect_list = info['intersect_list']
      polygon = MultiPolygon()
      for i in intersect_list:
        polygon = polygon.union(series.loc[i])
      intersect_info[index]['is_contains'] = polygon.contains(series.loc[index])
    return intersect_info

  def fix_intersects(geo_ser: pd.Series):
    intersect_info = ser_intersect(geo_ser)
    res_geo = []
    res_index = []
    contains_geo = []
    contains_index = []
    for index, geo in tqdm(geo_ser.items(),
                           desc='fix_intersects',
                           total=len(geo_ser)):
      info = intersect_info.get(index)
      intersect_list = info.get('intersect_list')
      if info.get('is_contains') and keep_contains:
        contains_geo.append(geo)
        contains_index.append(index)
      else:
        for i in intersect_list:
          if fill_intersects and i in res_index:
            continue
          geo = geo.difference(geo_ser.loc[i].buffer(1e-7))
        res_geo.append(filter_polygon_from_collection(geo))
        res_index.append(index)
    intersect_res = pd.Series(res_geo, index=res_index)
    if contains_index:
      if set(contains_index) == set(geo_ser.index):
        raise ValueError('mutual inclusion geometry')
      contains_res = fix_intersects(pd.Series(contains_geo,
                                              index=contains_index))
      return pd.concat([intersect_res, contains_res])
    else:
      return intersect_res

  assert series.is_unique, 'duplicated index'
  series = series.apply(ensure_valid_polygon)
  res = fix_intersects(series)
  return res


def fix_topology(df: gpd.GeoDataFrame,
                 c_geometry='geometry',
                 fill_intersects=False,
                 keep_contains=False) -> gpd.GeoDataFrame:
  """
  修复面地理数据的拓扑问题,index不能重复

  Args:
    df: 需要进行拓扑修复的GeoDataFrame
    c_geometry: geometry列列名
    fill_intersects: 是否填满相交的区域，如果为True的话，两个面相交的区域会随机分配到其中一个面内
    keep_contains: fill_intersects为True时才有效，即是否强制保留被包含的面

  Returns:
    修复完拓扑问题的GeoDataFrame
  """
  df = auto2shapely(df, geometry=c_geometry)
  if is_topology_valid(df, c_geometry):
    return df
  df[c_geometry] = _fix_topology(df[c_geometry],
                                 fill_intersects,
                                 keep_contains)
  return gpd.GeoDataFrame(df, crs=4326, geometry=c_geometry)
